/*
 *  sel2*, selview: several programs to view schedule event listings
 *  Copyright (C) 1998-2005  Ferdy Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <limits.h>

#include "actions.h"
#include "error.h"
#include "memalloc.h"
#include "sel.h"
#include "tasks.h"

static actiontree *find_next_free_node(
	actiontree *node,
	unsigned long time,
	int rsrc_id,
	action_type type)
{
	if (node == NULL) {
		return NULL;
	}
	if (time < node->start_time) {
		if (node->sooner == NULL) {
			return node;
		}
		return find_next_free_node(node->sooner, time, rsrc_id, type);
	} else {
		if (node->later == NULL) {
			return node;
		}
		return find_next_free_node(node->later, time, rsrc_id, type);
	}
}

actiontree *add_action(actiontree *at, action_type type, ...)
{
	actiontree *tmp;
	va_list ap;

	tmp = memalloc(sizeof(actiontree));
	tmp->type = type;
	tmp->start_time = 0;
	tmp->end_time = 0;
	tmp->rsrc_id = -1;
	tmp->new_deadline = 0;
	tmp->sooner = NULL;
	tmp->later = NULL;
	va_start(ap, type);
	switch (type) {
	case period:
	case exec:
		tmp->start_time = va_arg(ap, unsigned long);
		tmp->end_time = va_arg(ap, unsigned long);
		break;
	case resource_claim:
		tmp->rsrc_id = va_arg(ap, int);
		tmp->start_time = va_arg(ap, unsigned long);
		tmp->end_time = va_arg(ap, unsigned long);
		break;
	case deadline_redef:
		tmp->start_time = va_arg(ap, unsigned long);
		tmp->new_deadline = va_arg(ap, unsigned long);
		break;
	}
	va_end(ap);
	if (at == NULL) {
		at = tmp;
	} else {
		actiontree *node;

		node = find_next_free_node(at, tmp->start_time, tmp->rsrc_id, type);
		if (node) {
			if (tmp->start_time < node->start_time) {
				node->sooner = tmp;
			} else {
				node->later = tmp;
			}
		} else {
			selerror("cannot add action, something is terribly wrong");
		}
	}
	return at;
}

void free_actions(actiontree *at)
{
	if (at) {
		free_actions(at->sooner);
		free_actions(at->later);
		memfree(at);
	}
}

static unsigned long get_highest_time_from_action(actiontree *at, unsigned long high_until_now)
{
	unsigned long tmp;

	if (at) {
		switch (at->type) {
		case period:
		case exec:
		case resource_claim:
			if (at->start_time > high_until_now) {
				high_until_now = at->start_time;
			}
			if (at->end_time > high_until_now) {
				high_until_now = at->end_time;
			}
			break;
		case deadline_redef:
			if (at->new_deadline > high_until_now) {
				high_until_now = at->new_deadline;
			}
			break;
		}
		tmp = get_highest_time_from_action(at->sooner, high_until_now);
		if (tmp > high_until_now) {
			high_until_now = tmp;
		}
		tmp = get_highest_time_from_action(at->later, high_until_now);
		if (tmp > high_until_now) {
			high_until_now = tmp;
		}
	}
	return high_until_now;
}

unsigned long get_highest_time_from_actions(actiontree *at)
{
	return get_highest_time_from_action(at, 0);
}

static unsigned long get_lowest_time_from_action(actiontree *at, unsigned long low_until_now)
{
	unsigned long tmp;

	if (at) {
		switch (at->type) {
		case period:
		case exec:
		case resource_claim:
			if (at->start_time < low_until_now) {
				low_until_now = at->start_time;
			}
			if (at->end_time < low_until_now) {
				low_until_now = at->end_time;
			}
			break;
		case deadline_redef:
			if (at->new_deadline < low_until_now) {
				low_until_now = at->new_deadline;
			}
			break;
		}
		tmp = get_lowest_time_from_action(at->sooner, low_until_now);
		if (tmp < low_until_now) {
			low_until_now = tmp;
		}
		tmp = get_lowest_time_from_action(at->later, low_until_now);
		if (tmp < low_until_now) {
			low_until_now = tmp;
		}
	}
	return low_until_now;
}

unsigned long get_lowest_time_from_actions(actiontree *at)
{
	return get_lowest_time_from_action(at, ULONG_MAX);
}
