/*
 *  sel2*, selview: several programs to view schedule event listings
 *  Copyright (C) 1998-2005  Ferdy Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>

#include "actions.h"
#include "actions_num.h"
#include "memalloc.h"
#include "sel.h"
#include "tasks.h"

static int highest_rsrc_id_plus_one;

const char *actionlisttype2ascii(actionlist_type type)
{
	switch (type) {
	case release:
		return "release";
	case deadline:
		return "deadline";
	case exec_start:
		return "start of execution";
	case exec_end:
		return "end of execution";
	case claim_res:
		return "claim resource";
	case free_res:
		return "free resource";
	case redef_deadline:
		return "redefinition of deadline";
	default:
		return "unknown";
	}
}

static int before(
	actionlist *a,
	actionlist *b,
	int in_period,
	int in_exec,
	int *in_resource)
/* Returns true iff a comes before b in the action list */
{
	if (a->time < b->time) {
		return 1;
	}
	if (a->time > b->time) {
		return 0;
	}
	switch (a->type) {
	case release:
		switch (b->type) {
		case release:
			return 1;
		case deadline:
		case exec_start:
		case exec_end:
		case claim_res:
		case free_res:
		case redef_deadline:
			return !in_period;
		}
	case deadline:
		switch (b->type) {
		case release:
			return in_period;
		case deadline:
			return 1;
		case exec_start:
		case exec_end:
		case claim_res:
		case free_res:
		case redef_deadline:
			return !in_period;
		}
	case exec_start:
		switch (b->type) {
		case release:
		case deadline:
			return in_period;
		case exec_start:
			return 1;
		case exec_end:
		case claim_res:
		case free_res:
			return !in_exec;
		case redef_deadline:
			return 1;
		}
	case exec_end:
		switch (b->type) {
		case release:
		case deadline:
			return in_period;
		case exec_start:
			return in_exec;
		case exec_end:
			return 1;
		case claim_res:
		case free_res:
			return !in_exec;
		case redef_deadline:
			return 1;
		}
	case claim_res:
		switch (b->type) {
		case release:
		case deadline:
			return in_period;
		case exec_start:
		case exec_end:
			return in_exec;
		case claim_res:
			return (a->rsrc_id >= b->rsrc_id);
		case free_res:
			if (a->rsrc_id == b->rsrc_id) {
				return in_resource[a->rsrc_id];
			} else {
				return (a->rsrc_id < b->rsrc_id);
			}
		case redef_deadline:
			return 1;
		}
	case free_res:
		switch (b->type) {
		case release:
		case deadline:
			return in_period;
		case exec_start:
		case exec_end:
			return in_exec;
		case claim_res:
			if (a->rsrc_id == b->rsrc_id) {
				return in_resource[a->rsrc_id];
			} else {
				return (a->rsrc_id > b->rsrc_id);
			}
		case free_res:
			return (a->rsrc_id <= b->rsrc_id);
		case redef_deadline:
			return 1;
		}
	case redef_deadline:
		switch (b->type) {
		case release:
		case deadline:
			return in_period;
		case exec_start:
		case exec_end:
		case claim_res:
		case free_res:
			return 1;
		case redef_deadline:
			return (a->new_deadline < b->new_deadline);
		}
	}
	fprintf(stderr, "fatal: incomplete comparison -- should never happen");
	exit(1);
}

static actionlist *add2list(
	actionlist *al,
	actionlist_type type,
	unsigned long time,
	int rsrc_id,
	unsigned long new_deadline)
{
	actionlist *new, *next, *prev;
	int i, in_period, in_exec, *in_resource;

	new = memalloc(sizeof(actionlist));
	new->type = type;
	new->time = time;
	new->rsrc_id = rsrc_id;
	new->new_deadline = new_deadline;
	new->prev = NULL;
	new->next = NULL;
	in_period = 0;
	in_exec = 0;
	in_resource = memalloc(sizeof(int) * highest_rsrc_id_plus_one);
	for (i = 0; i < highest_rsrc_id_plus_one; i++) {
		in_resource[i] = 0;
	}
	for (prev = NULL, next = al; next; prev = next, next = next->next) {
		if (before(new, next, in_period, in_exec, in_resource)) {
			new->next = next;
			new->prev = prev;
			next->prev = new;
			if (prev) {
				/* add somewhere in the middle */
				prev->next = new;
			} else {
				/* add at the beginning of the list */
				al = new;
			}
			memfree(in_resource);
			return al;
		}
		switch(next->type) {
		case release:
			in_period = 1;
			break;
		case deadline:
			in_period = 0;
			break;
		case exec_start:
			in_exec = 1;
			break;
		case exec_end:
			in_exec = 0;
			break;
		case claim_res:
			in_resource[next->rsrc_id] = 1;
			break;
		case free_res:
			in_resource[next->rsrc_id] = 0;
			break;
		default:
			break;
		}
	}
	memfree(in_resource);
	if (prev) {
		/* add at end of list */
		new->prev = prev;
		prev->next = new;
		return al;
	} else {
		/* prev was empty => al was empty => list now has 1 element */
		return new;
	}
}

static actionlist *convert_tree_to_list_(actionlist *al, actiontree *at)
{
	if (at) {
		al = convert_tree_to_list_(al, at->sooner);
		switch (at->type) {
		case period:
			if (at->start_time <= maxtime) {
				al = add2list(al, release, at->start_time, 0, 0);
			}
			if (at->end_time <= maxtime) {
				al = add2list(al, deadline, at->end_time, 0, 0);
			}
			break;
		case exec:
			if (at->start_time <= maxtime) {
				al = add2list(al, exec_start, at->start_time, 0, 0);
			}
			if (at->end_time <= maxtime) {
				al = add2list(al, exec_end, at->end_time, 0, 0);
			}
			break;
		case resource_claim:
			if (at->start_time <= maxtime) {
				al = add2list(al, claim_res, at->start_time, at->rsrc_id, 0);
			}
			if (at->end_time <= maxtime) {
				al = add2list(al, free_res, at->end_time, at->rsrc_id, 0);
			}
			break;
		case deadline_redef:
			if (at->new_deadline <= maxtime) {
				al = add2list(al, redef_deadline, at->start_time, 0,
				              at->new_deadline);
			}
			break;
		}
		al = convert_tree_to_list_(al, at->later);
	}
	return al;
}

actionlist *convert_tree_to_list(actiontree *at)
{
	highest_rsrc_id_plus_one = get_highest_rsrc_id() + 1;
	return convert_tree_to_list_(NULL, at);
}

void free_actionlist(actionlist *al)
{
	if (al) {
		free_actionlist(al->next);
		memfree(al);
	}
}
