/*
 *  dat2avg: program to calculate averages from output of 'sel2num -d'
 *  Copyright (C) 1998-2005  Ferdy Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#define __USE_GNU    /* for isblank */
#define __USE_ISOC99 /* for isblank */
#include <ctype.h>

#include "doatexit.h"
#include "error.h"
#include "getopt.h"
#include "memalloc.h"
#include "strsave.h"
#include "version.h"

#ifdef PLUMBER
# include <plumber.h>
#endif

char *progname, *input_filename, *output_filename;
int sellineno;

void usage(void)
{
	fprintf(stderr, "Usage: %s [options] [input-file-name]\n" ,progname);
	fprintf(stderr, "Options:\n");
	fprintf(stderr, " -o, --output output-file-name\n");
	fprintf(stderr, " -c, --copyright\n");
	fprintf(stderr, " -V, --version\n");
	fprintf(stderr, " -h, --help\n");
	exit(1);
}

void version(void)
{
	printf("%s v" VERSION " by " AUTHOR "\n", progname);
	exit(0);
}

void copyright(void)
{
	printf("dat2avg V" VERSION "\n\n"
	       "dat2avg: program to calculate averages from output of 'sel2num -d'\n"
	       "Copyright (C) 1998-2005  Ferdy Hanssen <hanssen@croky.net>\n\n"
	       "This program is free software; you can redistribute it and/or modify\n"
	       "it under the terms of the GNU General Public License as published by\n"
	       "the Free Software Foundation; either version 2 of the License, or\n"
	       "(at your option) any later version.\n\n"
	       "This program is distributed in the hope that it will be useful,\n"
	       "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
	       "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
	       "GNU General Public License for more details.\n\n"
	       "You should have received a copy of the GNU General Public License\n"
	       "along with this program; if not, write to the Free Software\n"
	       "Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.\n");
	exit(0);
}

void process_options(unsigned int argc, char *argv[])
{
	char short_opts[] = "o:cVh";
	struct option long_opts[] = {
		{"output"         , required_argument, NULL, 'o'},
		{"copyright"      , no_argument      , NULL, 'c'},
		{"version"        , no_argument      , NULL, 'V'},
		{"help"           , no_argument      , NULL, 'h'},
		{0                , 0                , NULL, 0}
	};
	int c;

	output_filename = NULL;
	while ((c = getopt_long((int) argc, argv, short_opts, long_opts, NULL)) != EOF) {
		switch (c) {
		case 'o':
			output_filename = strsave(optarg);
			break;
		case 'c':
			copyright(); /* does not return */
		case 'V':
			version(); /* does not return */
		default: /* includes 'h' */
			usage(); /* does not return */
		}
	}
	if (argc - optind >= 2) { /* complain if more than 2 input files given */
		usage();
	}
	if (argc - optind == 0) { /* no input file given -> use stdin */
		input_filename = strsave("."); /* use . as code for stdin */
	} else {
		input_filename = strsave(argv[optind]);
	}
}

void cleanup_options(void)
{
	if (input_filename) {
		memfree(input_filename);
	}
	if (output_filename) {
		memfree(output_filename);
	}
}

int read_char(FILE *input)
{
	int c, comment;

	comment = 0;
	c = fgetc(input);
	while (comment || c == '#') {
		if (c == '#') {
			comment = 1;
		}
		c = fgetc(input);
		if (comment && c == '\n') {
			comment = 0;
		}
	}
	if (c == '\n') {
		sellineno++;
	}
	return c;
}

int unread_char(int c, FILE *input)
{
	if (c == '\n') {
		sellineno--;
	}
	return ungetc(c, input);
}

void skip_blanks(FILE *input)
{
	int c;

	c = read_char(input);
	while (c != EOF && (isblank(c) || c == '\n')) {
		c = read_char(input);
	}
	unread_char(c, input);
}

int read_number(unsigned long *number, FILE *input)
{
	int c;
	unsigned long num;

	num = 0;
	skip_blanks(input);
	c = read_char(input);
	if (c == EOF) {
		return -1;
	}
	while (isdigit(c)) {
		num = num * 10 + (c - '0');
		c = read_char(input);
	}
	unread_char(c, input);
	*number = num;
	return 0;
}

unsigned long expect_number(FILE *input, const char *expecting)
{
	int eof;
	unsigned long num;

	eof = read_number(&num, input);
	if (eof) {
		selerror("premature end of file, expecting %s", expecting);
		exit(1);
	}
	return num;
}

void process_file(FILE *input, FILE *out)
{
	unsigned long task, ntasks;

	sellineno = 0;
	ntasks = expect_number(input, "number of tasks");
	fprintf(out, "%lu tasks total\n", ntasks);
	for (task = 0; task < ntasks; task++) {
		unsigned long id, total_exec, total_delay, active_periods, periods;
		unsigned long resource, nresources;

		id = expect_number(input, "task id");
		total_exec = expect_number(input, "total exec time");
		total_delay = expect_number(input, "total time between release and exec");
		active_periods = expect_number(input, "number of active periods");
		periods = expect_number(input, "number of periods");
		nresources = expect_number(input, "number of resources used by this task");
		fprintf(out, "task %lu\n", id);
		fprintf(out, " avg. exec time %.2f\n",
		             (double) total_exec / (double) active_periods);
		fprintf(out, " avg. time between release and exec %.2f\n",
		             (double) total_delay / (double) active_periods);
		for (resource = 0; resource < nresources; resource++) {
			unsigned long id, total_usage, times_used;

			id = expect_number(input, "resource id");
			total_usage = expect_number(input, "total resource usage time");
			times_used = expect_number(input, "times resource used");
			fprintf(out, " resource %lu\n", id);
			fprintf(out, "  avg. usage time %.2f\n",
			             (double) total_usage / (double) times_used);
		}
	}
}

int main(int argc, char **argv)
{
	FILE *input, *output;

	progname = argv[0];
#ifdef PLUMBER
	plumber_init(progname);
	if (atexit(plumber_shutdown)) {
		selerror("cannot register plumber_shutdown() with atexit");
		plumber_shutdown();
		exit(1);
	}
#endif

	process_options(argc, argv);
	doatexit(cleanup_options);

	if (!strcmp(input_filename, ".")) {
		input = stdin;
	} else {
		if ((input = fopen(input_filename, "r")) == NULL || ferror(input)) {
			selerror("cannot open %s", input_filename);
			return 1;
		}
	}

	if (output_filename) {
		if ((output = fopen(output_filename, "w")) == NULL || ferror(output)) {
			selerror("cannot open %s", output_filename);
			fclose(input);
			return 1;
		}
	} else {
		output = stdout;
	}

	process_file(input, output);

	fclose(output);
	fclose(input);

	return 0;
}
