/*
 *  sel2*, selview: several programs to view schedule event listings
 *  Copyright (C) 1998-2005  Ferdy Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Supplementary functions for cropping of images,
 * for use in conjunction with the gd library
 *
 * Copyright (C) 1998-2005 by Ferdy Hanssen <hanssen@croky.net>
 *
 * Functions for cropping:
 * gdCropLeft(gdImagePtr *im, int background_color):   crops the left band of the
 *						       image <im>, which consists
 *						       of one color only: the
 *						       specified <background_color>
 * gdCropRight(gdImagePtr *im, int background_color):  crops the right band of the
 *						       image <im>, which consists
 *						       of one color only: the
 *						       specified <background_color>
 * gdCropTop(gdImagePtr *im, int background_color):    crops the upper band of the
 *						       image <im>, which consists
 *						       of one color only: the
 *						       specified <background_color>
 * gdCropBottom(gdImagePtr *im, int background_color): crops the lower band of the
 *						       image <im>, which consists
 *						       of one color only: the
 *						       specified <background_color>
 * gdCrop(gdImagePtr *im, int background_color):       combines gdCropLeft,
 *						       gdCropRight, gdCropTop, and
 *						       gdCropBottom
 */

#include <gd.h>

#include "gdCrop.h"

static void copy_region(gdImagePtr *im, int clx, int cly, int crx, int cry)
{
	gdImagePtr aux_im;

	aux_im = gdImageCreate(crx - clx, cry - cly);
	gdImageCopy(aux_im, *im, 0, 0, clx, cly, crx - clx, cry - cly);
	gdImageDestroy(*im);
	*im = aux_im;
}

void gdCropLeft(gdImagePtr *im, int background_color)
{
	register int x, y, tmp;
	int clx, cly, crx, cry;

	clx = 0;
	cly = 0;
	crx = gdImageSX(*im);
	cry = gdImageSY(*im);

	for (x = clx; x < crx; x++) { /* Crop left */
		tmp = 0;
		for (y = cly; !tmp && y < cry; y++) {
			if (gdImageGetPixel(*im, x, y) != background_color) {
				tmp = 1;
			}
		}
		if (tmp) {
			clx = x;
			break; /* out of for x */
		}
	}

	copy_region(im, clx, cly, crx, cry);
}

void gdCropRight(gdImagePtr *im, int background_color)
{
	register int x, y, tmp;
	int clx, cly, crx, cry;

	clx = 0;
	cly = 0;
	crx = gdImageSX(*im);
	cry = gdImageSY(*im);

	for (x = crx - 1; x >= clx; x--) { /* Crop right */
		tmp = 0;
		for (y = cly; !tmp && y < cry; y++) {
			if (gdImageGetPixel(*im, x, y) != background_color) {
				tmp = 1;
			}
		}
		if (tmp) {
			crx = x + 1;
			break; /* out of for x */
		}
	}

	copy_region(im, clx, cly, crx, cry);
}

void gdCropTop(gdImagePtr *im, int background_color)
{
	register int x, y, tmp;
	int clx, cly, crx, cry;

	clx = 0;
	cly = 0;
	crx = gdImageSX(*im);
	cry = gdImageSY(*im);

	for (y = cly; y < cry; y++) { /* Crop top */
		tmp = 0;
		for (x = clx; !tmp && x < crx; x++) {
			if (gdImageGetPixel(*im, x, y) != background_color) {
				tmp = 1;
			}
		}
		if (tmp) {
			cly = y;
			break; /* out of for y */
		}
	}

	copy_region(im, clx, cly, crx, cry);
}

void gdCropBottom(gdImagePtr *im, int background_color)
{
	register int x, y, tmp;
	int clx, cly, crx, cry;

	clx = 0;
	cly = 0;
	crx = gdImageSX(*im);
	cry = gdImageSY(*im);

	for (y = cry - 1; y >= cly; y--) { /* Crop bottom */
		tmp = 0;
		for (x = clx; !tmp && x < crx; x++) {
			if (gdImageGetPixel(*im, x, y) != background_color) {
				tmp = 1;
			}
		}
		if (tmp) {
			cry = y + 1;
			break; /* out of for y */
		}
	}

	copy_region(im, clx, cly, crx, cry);
}

void gdCrop(gdImagePtr *im, int background_color)
{
	register int x, y, tmp;
	int clx, cly, crx, cry;

	clx = 0;
	cly = 0;
	crx = gdImageSX(*im);
	cry = gdImageSY(*im);

	for (x = clx; x < crx; x++) { /* Crop left */
		tmp = 0;
		for (y = cly; !tmp && y < cry; y++) {
			if (gdImageGetPixel(*im, x, y) != background_color) {
				tmp = 1;
			}
		}
		if (tmp) {
			clx = x;
			break; /* out of for x */
		}
	}

	for (x = crx - 1; x >= clx; x--) { /* Crop right */
		tmp = 0;
		for (y = cly; !tmp && y < cry; y++) {
			if (gdImageGetPixel(*im, x, y) != background_color) {
				tmp = 1;
			}
		}
		if (tmp) {
			crx = x + 1;
			break; /* out of for x */
		}
	}

	for (y = cly; y < cry; y++) { /* Crop top */
		tmp = 0;
		for (x = clx; !tmp && x < crx; x++) {
			if (gdImageGetPixel(*im, x, y) != background_color) {
				tmp = 1;
			}
		}
		if (tmp) {
			cly = y;
			break; /* out of for y */
		}
	}

	for (y = cry - 1; y >= cly; y--) { /* Crop bottom */
		tmp = 0;
		for (x = clx; !tmp && x < crx; x++) {
			if (gdImageGetPixel(*im, x, y) != background_color) {
				tmp = 1;
			}
		}
		if (tmp) {
			cry = y + 1;
			break; /* out of for y */
		}
	}

	copy_region(im, clx, cly, crx, cry);
}
