/*
 *  sel2*, selview: several programs to view schedule event listings
 *  Copyright (C) 1998-2005  Ferdy Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Supplementary functions for creating XPMs
 * for use in conjunction with the gd library
 *
 * Copyright (C) 1998-2005 by Ferdy Hanssen <hanssen@croky.net>
 *
 * Functions:
 * char **gdCreateXpmDataFromImage(gdImagePtr im): creates a pixmap data array
 * void gdDestroyXpmData(char **data): destroys a pixmap data array
 * void gdImageXpm(gdImagePtr im, FILE *out): like gdImagePng, but creates an XPM
 */

#include <stdio.h>
#include <string.h>
#include <gd.h>

#include "gdImageXpm.h"
#include "memalloc.h"
#include "strsave.h"

char **gdCreateXpmDataFromImage(gdImagePtr im)
{
	char **xpm;
	char *tmp;
	char colmaps[256][2];
	char colchars[] = ".#abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
	                  "!@$%^&*()-=_+[]{};:',/<>?|`~";
	int i, x, y, width, height, num_colors, trans_color, cpp;

#ifdef DEBUG
	fprintf(stderr, "gdCreateDataFromImage: start\n");
#endif

	width = gdImageSX(im);
	height = gdImageSY(im);
	num_colors = gdImageColorsTotal(im);
	trans_color = gdImageGetTransparent(im);
	if ((size_t) num_colors > strlen(colchars)) {
		cpp = 2; /* plenty for 256 colors */
	} else {
		cpp = 1;
	}

	tmp = memalloc(sizeof(char) * 256);

	xpm = memalloc(sizeof(char *) * (1 + num_colors + height));
	snprintf(tmp, 256, "%d %d %d %d", width, height, num_colors, cpp);
	xpm[0] = strsave(tmp);

	if (cpp == 1) {
		for (i = 0; i < num_colors; i++) {
			if (i == trans_color) {
				snprintf(tmp, 256, "%c c None", colchars[i]);
			} else {
				snprintf(tmp, 256, "%c c #%02x%02x%02x", colchars[i],
				                   gdImageRed(im, i), gdImageGreen(im, i),
				                   gdImageBlue(im ,i));
			}
			colmaps[i][0] = colchars[i];
			xpm[i + 1] = strsave(tmp);
		}
	} else {
		int p, q;

		p = 0;
		q = 0;
		for (i = 0; i < num_colors; i++) {
			if (i == trans_color) {
				snprintf(tmp, 256, "%c%c c None", colchars[p], colchars[q]);
			} else {
				snprintf(tmp, 256, "%c%c c #%02x%02x%02x", colchars[p],
				                   colchars[q], gdImageRed(im, i),
				                   gdImageGreen(im, i), gdImageBlue(im ,i));
			}
			xpm[i + 1] = strsave(tmp);
			colmaps[i][0] = colchars[p];
			colmaps[i][1] = colchars[q];
			p++;
			if ((size_t) p == strlen(colchars)) {
				p = 0;
				q++;
			}
		}
	}

	memfree(tmp);
	tmp = memalloc(sizeof(char) * (width * cpp + 1));
	tmp[width * cpp] = '\0';
	if (cpp == 1) {
		for (y = 0; y < height; y++) {
			for (x = 0; x < width ; x++) {
				tmp[x] = colmaps[gdImageGetPixel(im, x, y)][0];
			}
			xpm[y + num_colors + 1] = strsave(tmp);
		}
	} else {
		for (y = 0; y < height; y++) {
			for (x = 0; x < width ; x++) {
				tmp[x * 2] = colmaps[gdImageGetPixel(im, x, y)][0];
				tmp[x * 2 + 1] = colmaps[gdImageGetPixel(im, x, y)][1];
			}
			xpm[y + num_colors + 1] = strsave(tmp);
		}
	}

	memfree(tmp);

#ifdef DEBUG
	fprintf(stderr, "gdCreateDataFromImage: end\n");
#endif

	return xpm;
}

void gdDestroyXpmData(char **data)
{
	int i, height, num_colors;

	if (sscanf(*data, "%*d %d %d", &height, &num_colors) != 2) {
		fprintf(stderr, "WARNING: gdDestroyXpmData: cannot parse pixmap data array at "
		                "%p\n", data);
		return;
	}

	for (i = num_colors + height; i >= 0; i--) {
		memfree(data[i]);
	}
	memfree(data);
}

void gdImageXpm(gdImagePtr im, FILE *out)
{
	char **xpm;
	int i, height, num_colors;

#ifdef DEBUG
	fprintf(stderr, "gdImageXpm: start\n");
#endif

	xpm = gdCreateXpmDataFromImage(im);

	height = gdImageSY(im);
	num_colors = gdImageColorsTotal(im);

	fprintf(out, "/* XPM */\n");
	fprintf(out, "static char *xpm_image[] = {\n");
	for (i = 0; i < 1 + num_colors + height; i++) {
		fprintf(out, "\"%s\",\n", xpm[i]);
	}
	fprintf(out, "};\n");

	gdDestroyXpmData(xpm);

#ifdef DEBUG
	fprintf(stderr, "gdImageXpm: end\n");
#endif
}
