/*
 *  sel2*, selview: several programs to view schedule event listings
 *  Copyright (C) 1998-2005  Ferdy Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <gtk/gtk.h>

#ifndef CHECK_UNUSED
# ifdef DEBUG
#  define unused(x)	(x) = (x)
# else
#  define unused(x)
# endif
#else
# define unused(x)
#endif
#ifdef HIDE_UNUSED
# undef unused
# define unused(x)	(x) = (x)
#endif

#include "gtknumberrange.h"
#include "gtkpreviewhelper.h"

#include "up.h"
#include "down.h"
#include "bigup.h"
#include "bigdown.h"

#define TEXT_FIELD_LENGTH	7
#define TEXT_FIELD_WIDTH	(9 + TEXT_FIELD_LENGTH * 7)
#define TEXT_FIELD_HEIGHT	22

static void ensure_correctness(GtkNumberrange *w)
{
	if (w->value > w->maximum) {
		w->value = w->maximum;
	}
	if (w->value < w->minimum) {
		w->value = w->minimum;
	}
}

static void set_text(GtkNumberrange *w)
{
	char s[13];

	snprintf(s, 13, "%lu", w->value);
	gtk_entry_set_text(GTK_ENTRY(w->numberfield), s);
}

static void upbutton_pressed(GtkWidget *nf, GtkNumberrange *w)
{
	unused(nf);

	gtk_entry_set_editable(GTK_ENTRY(w->numberfield), FALSE);
	if (w->value + w->change < w->maximum) {
		w->value += w->change;
	} else {
		w->value = w->maximum;
	}
	set_text(w);
	gtk_entry_set_editable(GTK_ENTRY(w->numberfield), TRUE);
	w->action(w, w->value);
}

static void downbutton_pressed(GtkWidget *nf, GtkNumberrange *w)
{
	unused(nf);

	gtk_entry_set_editable(GTK_ENTRY(w->numberfield), FALSE);
	if (w->value > w->minimum + w->change) {
		w->value -= w->change;
	} else {
		w->value = w->minimum;
	}
	set_text(w);
	gtk_entry_set_editable(GTK_ENTRY(w->numberfield), TRUE);
	w->action(w, w->value);
}

static void bigupbutton_pressed(GtkWidget *nf, GtkNumberrange *w)
{
	unused(nf);

	gtk_entry_set_editable(GTK_ENTRY(w->numberfield), FALSE);
	if (w->value + w->big_change < w->maximum) {
		w->value += w->big_change;
	} else {
		w->value = w->maximum;
	}
	set_text(w);
	gtk_entry_set_editable(GTK_ENTRY(w->numberfield), TRUE);
	w->action(w, w->value);
}

static void bigdownbutton_pressed(GtkWidget *nf, GtkNumberrange *w)
{
	unused(nf);

	gtk_entry_set_editable(GTK_ENTRY(w->numberfield), FALSE);
	if (w->value > w->minimum + w->big_change) {
		w->value -= w->big_change;
	} else {
		w->value = w->minimum;
	}
	set_text(w);
	gtk_entry_set_editable(GTK_ENTRY(w->numberfield), TRUE);
	w->action(w, w->value);
}

static void numberfield_enter(GtkWidget *nf, GtkNumberrange *w)
{
	unsigned long n;
	char *s, *end;

	unused(nf);

	gtk_entry_set_editable(GTK_ENTRY(w->numberfield), FALSE);
	s = gtk_entry_get_text(GTK_ENTRY(w->numberfield));
	n = strtoul(s, &end, 10);
	if (*end == '\0' && *s != '\0') {
		w->value = n;
	}
	ensure_correctness(w);
	set_text(w);
	gtk_entry_set_editable(GTK_ENTRY(w->numberfield), TRUE);
	w->action(w, w->value);
}

GtkWidget *gtk_numberrange_new(
	const char *label,
	unsigned long initial_value,
	unsigned long minimum,
	unsigned long maximum,
	unsigned long change,
	unsigned long big_change,
	void (*action)(GtkNumberrange *w, unsigned long new_value))
{
	GtkNumberrange *w;
	GtkTable *tmp_table;
	GtkWidget *labelwid, *up, *down, *bigup, *bigdown;
	char s[13];

	if (minimum > maximum) {
		long tmp;

		tmp = minimum;
		minimum = maximum;
		maximum = tmp;
	}

	w = g_malloc(sizeof(GtkNumberrange));
	w->value = initial_value;
	w->minimum = minimum;
	w->maximum = maximum;
	w->change = change;
	w->big_change = big_change;
	w->action = action;
	ensure_correctness(w);

	tmp_table = (GtkTable *)gtk_table_new(1, 4, FALSE);
	memmove(&w->table, tmp_table, sizeof(GtkTable));

	labelwid = gtk_label_new(label);
	gtk_table_attach(GTK_TABLE(&w->table), labelwid, 0, 1, 0, 1, 0, 0, 0, 0);
	gtk_widget_show(labelwid);

	w->numberfield = gtk_entry_new_with_max_length(TEXT_FIELD_LENGTH);
	gtk_widget_set_usize(GTK_WIDGET(w->numberfield), TEXT_FIELD_WIDTH, TEXT_FIELD_HEIGHT);
	snprintf(s, 13, "%lu", initial_value);
	gtk_signal_connect(GTK_OBJECT(w->numberfield), "activate",
	                   GTK_SIGNAL_FUNC(numberfield_enter), w);
	gtk_entry_set_text(GTK_ENTRY(w->numberfield), s);
	gtk_table_attach(GTK_TABLE(&w->table), w->numberfield, 1, 2, 0, 1, 0, 0, 0, 0);
	gtk_widget_show(w->numberfield);

	w->bigupbutton = gtk_button_new();
	gtk_signal_connect(GTK_OBJECT(w->bigupbutton), "clicked",
	                   GTK_SIGNAL_FUNC(bigupbutton_pressed), w);
	bigup = gtk_build_preview((guchar *) bigup_data, (gint) bigup_width, (gint) bigup_height);
	gtk_container_add(GTK_CONTAINER(w->bigupbutton), bigup);
	gtk_widget_show(bigup);
	gtk_table_attach(GTK_TABLE(&w->table), w->bigupbutton, 2, 3, 0, 1, 0, 0, 0, 0);
	gtk_widget_show(w->bigupbutton);

	w->upbutton = gtk_button_new();
	gtk_signal_connect(GTK_OBJECT(w->upbutton), "clicked",
	                   GTK_SIGNAL_FUNC(upbutton_pressed), w);
	up = gtk_build_preview((guchar *) up_data, (gint) up_width, (gint) up_height);
	gtk_container_add(GTK_CONTAINER(w->upbutton), up);
	gtk_widget_show(up);
	gtk_table_attach(GTK_TABLE(&w->table), w->upbutton, 3, 4, 0, 1, 0, 0, 0, 0);
	gtk_widget_show(w->upbutton);

	w->downbutton = gtk_button_new();
	gtk_signal_connect(GTK_OBJECT(w->downbutton), "clicked",
	                   GTK_SIGNAL_FUNC(downbutton_pressed), w);
	down = gtk_build_preview((guchar *) down_data, (gint) down_width, (gint) down_height);
	gtk_container_add(GTK_CONTAINER(w->downbutton), down);
	gtk_widget_show(down);
	gtk_table_attach(GTK_TABLE(&w->table), w->downbutton, 4, 5, 0, 1, 0, 0, 0, 0);
	gtk_widget_show(w->downbutton);

	w->bigdownbutton = gtk_button_new();
	gtk_signal_connect(GTK_OBJECT(w->bigdownbutton), "clicked",
	                   GTK_SIGNAL_FUNC(bigdownbutton_pressed), w);
	bigdown = gtk_build_preview((guchar *) bigdown_data, (gint) bigdown_width, (gint) bigdown_height);
	gtk_container_add(GTK_CONTAINER(w->bigdownbutton), bigdown);
	gtk_widget_show(bigdown);
	gtk_table_attach(GTK_TABLE(&w->table), w->bigdownbutton, 5, 6, 0, 1, 0, 0, 0, 0);
	gtk_widget_show(w->bigdownbutton);

	return (GtkWidget *)w;
}

void gtk_numberrange_destroy(GtkNumberrange *w)
{
	gtk_widget_destroy(w->numberfield);
	gtk_widget_destroy(w->upbutton);
	gtk_widget_destroy(w->downbutton);
	g_free(w);
}

void gtk_numberrange_set(GtkNumberrange *w, unsigned long value)
{
	w->value = value;
	ensure_correctness(w);
	set_text(w);
}

void gtk_numberrange_minimum_set(GtkNumberrange *w, unsigned long minimum)
{
	w->minimum = minimum;
	ensure_correctness(w);
	set_text(w);
}

void gtk_numberrange_maximum_set(GtkNumberrange *w, unsigned long maximum)
{
	w->maximum = maximum;
	ensure_correctness(w);
	set_text(w);
}

void gtk_numberrange_change_set(GtkNumberrange *w, unsigned long change)
{
	w->change = change;
}

void gtk_numberrange_big_change_set(GtkNumberrange *w, unsigned long big_change)
{
	w->big_change = big_change;
}

unsigned long gtk_numberrange_get(GtkNumberrange *w)
{
	return w->value;
}

unsigned long gtk_numberrange_minimum_get(GtkNumberrange *w)
{
	return w->minimum;
}

unsigned long gtk_numberrange_maximum_get(GtkNumberrange *w)
{
	return w->maximum;
}

unsigned long gtk_numberrange_change_get(GtkNumberrange *w)
{
	return w->change;
}

unsigned long gtk_numberrange_big_change_get(GtkNumberrange *w)
{
	return w->big_change;
}
