/*
 *  sel2*, selview: several programs to view schedule event listings
 *  Copyright (C) 1998-2005  Ferdy Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>

#include <gd.h>

#include <gtk/gtk.h>

#include "doatexit.h"
#include "error.h"
#include "gtknumberrange.h"
#include "gtkpreviewhelper.h"
#include "gui.h"
#include "guicallbacks.h"
#include "guimenufactory.h"
#include "guishow.h"
#include "guistatus.h"
#include "input.h"
#include "options.h"
#include "output_graph.h"
#include "sel.h"
#include "sel.parser.h"
#include "tasks.h"

#ifdef PLUMBER
# include <plumber.h>
#endif

#include "leftarrow.h"
#include "rightarrow.h"

#include "empty.xpm"
#include "selview.xpm"

static void quit(int signal)
{
	sellineno = 0;
	switch (signal) {
	case SIGBUS:
		selerror("bus error");
		break;
	case SIGSEGV:
		selerror("segmentation fault");
		break;
#ifdef DEBUG
	default:
		selerror("caught signal: %s", g_strsignal(signal));
		break;
#endif
	}
	fflush(NULL);
#ifdef PLUMBER
	cleanup_options();
	plumber_shutdown();
#endif
	/* sadly fcloseall isn't generally available -- let's hope all goes well */
	_exit(1);
}

static void do_signal(int signum, void (*handler)(int))
{
	if (signal(signum, handler) == SIG_ERR) {
		selerror("cannot set signal handler for %s", g_strsignal(signum));
	}
}

int main(int argc, char *argv[])
{
	GtkWidget *main_vbox, *buttonbar, *menubar, *scrolled_window;
	GtkWidget *fixedinterval_button, *fixedinterval_button_box;
	GtkWidget *leftarrow, *rightarrow;
	GtkStyle *style;

	progname = argv[0];
#ifdef PLUMBER
	plumber_init(progname);
	if (atexit(plumber_shutdown)) {
		selerror("cannot register plumber_shutdown() with atexit");
		plumber_shutdown();
		exit(1);
	}
#endif

	process_options(argc, argv);

	gtk_init(&argc, &argv);

	do_signal(SIGBUS, quit);
	do_signal(SIGHUP, quit);
	do_signal(SIGINT, quit);
	do_signal(SIGPIPE, quit);
	do_signal(SIGQUIT, quit);
	do_signal(SIGSEGV, quit);
	do_signal(SIGTERM, quit);

	gui_error_init();

	doatexit(cleanup_options);

	doatexit(free_resources);
	doatexit(free_tasks);

	timemin = 0;
	timemax = MAXTIMEMAX;

	main_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_signal_connect(GTK_OBJECT(main_window), "delete_event",
	                   GTK_SIGNAL_FUNC(file_quit_cmd_callback), FALSE);

	gtk_window_set_title(GTK_WINDOW(main_window), "selview");
	gtk_widget_set_usize(GTK_WIDGET(main_window), window_width, window_height);

	gtk_widget_realize(main_window);
	style = gtk_widget_get_style(main_window);

	icon = gdk_pixmap_create_from_xpm_d(main_window->window, NULL,
	                                    &style->bg[GTK_STATE_NORMAL], (char **)selview);
	gdk_window_set_icon(main_window->window, NULL, icon, NULL);
	gdk_window_set_icon_name(main_window->window, "selview");

	main_vbox = gtk_vbox_new(FALSE, 1);
	gtk_container_set_border_width(GTK_CONTAINER(main_vbox), 1);
	gtk_container_add(GTK_CONTAINER(main_window), main_vbox);
	gtk_widget_show(main_vbox);

	get_main_menu(main_window, &menubar);
	gtk_box_pack_start(GTK_BOX(main_vbox), menubar, FALSE, FALSE, 0);
	gtk_widget_show(menubar);

	buttonbar = gtk_hbox_new(FALSE, 5);

	guizoom = gtk_numberrange_new("Zoom: ", zoom, 1, zoommax, 1, 2, zoom_change_callback);
	gtk_box_pack_start(GTK_BOX(buttonbar), guizoom, FALSE, FALSE, 1);
	gtk_widget_show(guizoom);

	guimintime = gtk_numberrange_new("Lower limit: ", mintime, timemin, timemax - 50, 50, 1000,
	                                 mintime_change_callback);
	gtk_box_pack_start(GTK_BOX(buttonbar), guimintime, FALSE, FALSE, 1);
	gtk_widget_show(guimintime);

	leftarrow = gtk_build_preview((guchar *) leftarrow_data, (gint) leftarrow_width, (gint) leftarrow_height);
	gtk_box_pack_start(GTK_BOX(buttonbar), leftarrow, FALSE, FALSE, 0);
	gtk_widget_show(leftarrow);

	fixedinterval_button = gtk_toggle_button_new();
	gtk_signal_connect(GTK_OBJECT(fixedinterval_button), "toggled",
	                   GTK_SIGNAL_FUNC(fixedinterval_change_callback), NULL);
	gtk_box_pack_start(GTK_BOX(buttonbar), fixedinterval_button, FALSE, FALSE, 3);
	fixedinterval_button_box = gtk_hbox_new(FALSE, 2);

	fixedinterval_label = gtk_label_new("variable");
	gtk_widget_show(fixedinterval_label);
	gtk_box_pack_start(GTK_BOX(fixedinterval_button_box), fixedinterval_label, TRUE, FALSE, 2);

	gtk_container_add(GTK_CONTAINER(fixedinterval_button), fixedinterval_button_box);
	gtk_widget_show(fixedinterval_button_box);

	gtk_widget_show(fixedinterval_button);

	rightarrow = gtk_build_preview((guchar *) rightarrow_data, (gint) rightarrow_width, (gint) rightarrow_height);
	gtk_box_pack_start(GTK_BOX(buttonbar), rightarrow, FALSE, FALSE, 0);
	gtk_widget_show(rightarrow);

	guimaxtime = gtk_numberrange_new("Upper limit: ", maxtime, timemin + 50, timemax, 50, 1000,
	                                 maxtime_change_callback);
	gtk_box_pack_start(GTK_BOX(buttonbar), guimaxtime, FALSE, FALSE, 1);
	gtk_widget_show(guimaxtime);

	gtk_widget_show(buttonbar);
	gtk_box_pack_start(GTK_BOX(main_vbox), buttonbar, FALSE, FALSE, 0);

	pixmap = gdk_pixmap_create_from_xpm_d(main_window->window, &mask,
	                                      &style->bg[GTK_STATE_NORMAL], (char **)empty);
	pixmapwid = gtk_pixmap_new(pixmap, mask);
	gtk_widget_show(pixmapwid);

	scrolled_window = gtk_scrolled_window_new(NULL, NULL);
	gtk_container_set_border_width(GTK_CONTAINER(scrolled_window), 2);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window),
	                               GTK_POLICY_ALWAYS, GTK_POLICY_AUTOMATIC);

	gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(scrolled_window), pixmapwid);
	gtk_container_add(GTK_CONTAINER(main_vbox), scrolled_window);
	gtk_widget_show(scrolled_window);

	statusbar = gtk_statusbar_new();
	statusbar_context_id = gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar), "selview");
	gtk_widget_show(statusbar);
	gtk_box_pack_end(GTK_BOX(main_vbox), statusbar, FALSE, FALSE, 0);

	gtk_widget_show(main_window);

	gtk_widget_set_usize(fixedinterval_label, fixedinterval_label->allocation.width, fixedinterval_label->allocation.height);

	if (fixedinterval) {
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(fixedinterval_button), TRUE);
	}

	push_msg("Ready."); /* Default message, will never be popped. */

	if (strcmp(input_filename, ".")) {
		load_file();
	}

	gtk_main();

	return 0;
}
