/*
 *  sel2*, selview: several programs to view schedule event listings
 *  Copyright (C) 1998-2005  Ferdy Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <unistd.h>
#include <pwd.h>
#include <sys/types.h>

#include "output_epsgraph.h"

/* length of string to use for creating numbers, hopefully long enough */
#define NUMSTRINGLEN 64

#include "actions.h"
#include "epslibrary.h"
#include "error.h"
#include "memalloc.h"
#include "sel.h"
#include "strcatsave.h"
#include "strsave.h"
#include "tasks.h"
#include "version.h"

extern tasktree *tasks;

struct elem1tuple {
	int64 e;
	struct elem1tuple *next;
};

typedef struct onetuplelist {
	struct elem1tuple *first;
	struct elem1tuple *last;
} onetuplelist;

struct elem2tuple {
	int64 e[2];
	struct elem2tuple *next;
};

typedef struct twotuplelist {
	struct elem2tuple *first;
	struct elem2tuple *last;
} twotuplelist;

struct elem3tuple {
	int64 e[3];
	struct elem3tuple *next;
};

typedef struct threetuplelist {
	struct elem3tuple *first;
	struct elem3tuple *last;
} threetuplelist;

typedef struct imagelist {
	twotuplelist exec;
	twotuplelist acti;
	threetuplelist resources;
	onetuplelist rdfd_deadlines;
	char *name;
	struct imagelist *next;
} imagelist;

static void add_onetuple(onetuplelist *val, int64 e0)
{
	struct elem1tuple *elem;

	elem = memalloc(sizeof(struct elem1tuple));
	elem->e = e0;
	elem->next = NULL;
	if (val->first) {
		val->last->next = elem;
	} else {
		val->first = elem;
	}
	val->last = elem;
}

static void free_onetuplelist(onetuplelist val)
{
	struct elem1tuple *elem, *next;

	elem = val.first;
	while (elem) {
		next = elem->next;
		memfree(elem);
		elem = next;
	}
}

static char *output_onetuplelist(onetuplelist val)
{
	struct elem1tuple *elem;
	char buf[NUMSTRINGLEN];
	char *output;

	elem = val.first;
	output = strsave("[");
	while (elem) {
		snprintf(buf, NUMSTRINGLEN, "%lli", elem->e);
		strcatsave(&output, buf);
		if (elem->next) {
			strcatsave(&output, " ");
		}
		elem = elem->next;
	}
	strcatsave(&output, "]");
	return output;
}

static void add_twotuple(twotuplelist *val, int64 e0, int64 e1)
{
	struct elem2tuple *elem;

	elem = memalloc(sizeof(struct elem2tuple));
	elem->e[0] = e0;
	elem->e[1] = e1;
	elem->next = NULL;
	if (val->first) {
		val->last->next = elem;
	} else {
		val->first = elem;
	}
	val->last = elem;
}

static void free_twotuplelist(twotuplelist val)
{
	struct elem2tuple *elem, *next;

	elem = val.first;
	while (elem) {
		next = elem->next;
		memfree(elem);
		elem = next;
	}
}

static char *output_twotuplelist(twotuplelist val)
{
	struct elem2tuple *elem;
	char buf[NUMSTRINGLEN];
	char *output;

	elem = val.first;
	output = strsave("[");
	while (elem) {
		strcatsave(&output, "[");
		snprintf(buf, NUMSTRINGLEN, "%lli", elem->e[0]);
		strcatsave(&output, buf);
		strcatsave(&output, " ");
		snprintf(buf, NUMSTRINGLEN, "%lli", elem->e[1]);
		strcatsave(&output, buf);
		strcatsave(&output, "]");
		elem = elem->next;
	}
	strcatsave(&output, "]");
	return output;
}

static void add_threetuple(threetuplelist *val, int64 e0, int64 e1, int64 e2)
{
	struct elem3tuple *elem;

	elem = memalloc(sizeof(struct elem3tuple));
	elem->e[0] = e0;
	elem->e[1] = e1;
	elem->e[2] = e2;
	elem->next = NULL;
	if (val->first) {
		val->last->next = elem;
	} else {
		val->first = elem;
	}
	val->last = elem;
}

static void free_threetuplelist(threetuplelist val)
{
	struct elem3tuple *elem, *next;

	elem = val.first;
	while (elem) {
		next = elem->next;
		memfree(elem);
		elem = next;
	}
}

static char *output_threetuplelist(threetuplelist val)
{
	struct elem3tuple *elem;
	char buf[NUMSTRINGLEN];
	char *output;

	elem = val.first;
	output = strsave("[");
	while (elem) {
		strcatsave(&output, "[");
		snprintf(buf, NUMSTRINGLEN, "%lli", elem->e[0]);
		strcatsave(&output, buf);
		strcatsave(&output, " ");
		snprintf(buf, NUMSTRINGLEN, "%lli", elem->e[1]);
		strcatsave(&output, buf);
		strcatsave(&output, " ");
		snprintf(buf, NUMSTRINGLEN, "%lli", elem->e[2]);
		strcatsave(&output, buf);
		strcatsave(&output, "]");
		elem = elem->next;
	}
	strcatsave(&output, "]");
	return output;
}

static void free_imagelist(imagelist *il)
{
	if (il) {
		free_imagelist(il->next);
		free_twotuplelist(il->exec);
		free_twotuplelist(il->acti);
		free_threetuplelist(il->resources);
		free_onetuplelist(il->rdfd_deadlines);
		memfree(il);
	}
}

static char *char2pschar(char c)
{
	static char buf[5];

	switch (c) {
	case '(': return "\\(";
	case ')': return "\\)";
	case '\\': return "\\";
	case '\n': return "\\n";
	case '\r': return "\\r";
	case '\t': return "\\t";
	default:
		if (c < ' ' || c >= '\177') {
			snprintf(buf, 5, "\%03o", (unsigned int) c);
		} else {
			snprintf(buf, 2, "%c", c);
		}
		return buf;
	}
}

static char *str2psstr(const char *s)
{
	char *result;

	result = strsave("(");
	while (*s != '\0') {
		strcatsave(&result, char2pschar(*s));
		s++;
	}
	strcatsave(&result, ")");
	return result;
}

static char *num2psstr(long n)
{
	static char buf[NUMSTRINGLEN];

	snprintf(buf, NUMSTRINGLEN, "%li", n);
	return buf;
}

static void make_list_from_actions(
	actiontree *at,
	unsigned long minimagetime,
	unsigned long maximagetime,
	twotuplelist *execs,
	twotuplelist *actis,
	threetuplelist *resources,
	onetuplelist *rdfd_deadlines)
{
	int64 start, end;

	if (at) {
		make_list_from_actions(at->sooner, minimagetime, maximagetime, execs, actis,
		                       resources, rdfd_deadlines);
		start = at->start_time;
		end = at->end_time;
		switch (at->type) {
		case period:
			if (end < minimagetime || start > maximagetime) {
				break;
			}
			if (start < minimagetime || start > maximagetime) {
				start = -1;
			}
			if (end < minimagetime || end > maximagetime) {
				end = -1;
			}
			add_twotuple(actis, start, end);
			break;
		case exec:
			if (end < minimagetime || start > maximagetime) {
				break;
			}
			if (start < minimagetime) {
				start = minimagetime;
			}
			if (end > maximagetime) {
				end = maximagetime;
			}
			add_twotuple(execs, start, end);
			break;
		case resource_claim:
			if (end < minimagetime || start > maximagetime) {
				break;
			}
			if (start < minimagetime) {
				start = minimagetime;
			}
			if (end > maximagetime) {
				end = maximagetime;
			}
			add_threetuple(resources, start, end, at->rsrc_id);
			break;
		case deadline_redef:
			if (at->new_deadline >= minimagetime &&
			    at->new_deadline <= maximagetime) {
				add_onetuple(rdfd_deadlines, at->new_deadline);
			}
			break;
		}
		make_list_from_actions(at->later, minimagetime, maximagetime, execs, actis,
		                       resources, rdfd_deadlines);
	}
}

static imagelist *make_image_from_task(
	tasktree *task,
	unsigned long minimagetime,
	unsigned long maximagetime)
{
	imagelist *result, *leftlast, *tmp;
	twotuplelist exec = { NULL, NULL }, acti = { NULL, NULL };
	threetuplelist resources = { NULL, NULL };
	onetuplelist rdfd_deadlines = { NULL, NULL };

	result = NULL;
	if (task) {
		/* get left part */
		result = make_image_from_task(task->left, minimagetime, maximagetime);

		/* get my image */
		make_list_from_actions(task->actions, minimagetime, maximagetime, &exec, &acti,
		                       &resources, &rdfd_deadlines);
		tmp = memalloc(sizeof(imagelist));
		tmp->exec = exec;
		tmp->acti = acti;
		tmp->resources = resources;
		tmp->rdfd_deadlines = rdfd_deadlines;
		tmp->name = task->name;
		tmp->next = NULL;

		if (result) {
			/* search for last element of left part */
			leftlast = result;
			while (leftlast->next) {
				leftlast = leftlast->next;
			}

			/* put my image at the back of the left part */
			leftlast->next = tmp;
			leftlast = leftlast->next;
		} else {
			result = leftlast = tmp;
		}

		/* get right part */
		tmp = make_image_from_task(task->right, minimagetime, maximagetime);

		/* concatenate left and right parts */
		leftlast->next = tmp;
	}
	return result;
}

char *build_eps(const char *title)
{
	int namefieldlen, bbl, bbb, bbr, bbt, number;
	unsigned long minimagetime, maximagetime;
	imagelist *schedules, *tmp;
	char *result, *schedulestr, *output;
	struct passwd *pe;

#ifdef DEBUG
	fprintf(stderr, "build_eps: start\n");
#endif

	namefieldlen = get_longest_name_length_from_tasks();
	if (namefieldlen < 5) { /* 5 is the length of the word Tasks */
		namefieldlen = 5;
	}
	minimagetime = (get_lowest_action_time_from_tasks() / 50) * 50;
	if (minimagetime < mintime) {
		minimagetime = (mintime / 5) * 5;
	} else if (minimagetime > maxtime) {
		minimagetime = (maxtime / 5) * 5;
	}
	maximagetime = ((get_highest_action_time_from_tasks() + 49) / 50) * 50;
	if (maximagetime > maxtime) {
		maximagetime = ((maxtime + 4) / 5) * 5;
	} else if (maximagetime < mintime) {
		maximagetime = ((mintime + 4) / 5) * 5;
	}
	if (maximagetime <= minimagetime) {
		maximagetime = minimagetime + 50;
	}
	{
		char s[NUMSTRINGLEN];

		if (snprintf(s, NUMSTRINGLEN, "%lu", maximagetime) == -1) {
			selerror("highest action time too high to produce output");
			return NULL;
		}
	}

	/* make schedules */
	schedules = make_image_from_task(tasks, minimagetime, maximagetime);

	/* generate schedules */
	schedulestr = strsave("");
	tmp = schedules;
	number = 0;
	while (tmp) {
		number++;

		output = output_onetuplelist(tmp->rdfd_deadlines);
		strcatsave(&schedulestr, output);
		memfree(output);
		strcatsave(&schedulestr, " ");

		output = output_twotuplelist(tmp->acti);
		strcatsave(&schedulestr, output);
		memfree(output);
		strcatsave(&schedulestr, " ");

		output = str2psstr(tmp->name);
		strcatsave(&schedulestr, output);
		memfree(output);
		strcatsave(&schedulestr, " ");
		strcatsave(&schedulestr, num2psstr(maximagetime));
		strcatsave(&schedulestr, " ");
		if (tmp->next) {
			strcatsave(&schedulestr, "false ");
		} else {
			strcatsave(&schedulestr, "true ");
		}

		output = output_threetuplelist(tmp->resources);
		strcatsave(&schedulestr, output);
		memfree(output);
		strcatsave(&schedulestr, " ");

		output = output_twotuplelist(tmp->exec);
		strcatsave(&schedulestr, output);
		memfree(output);
		strcatsave(&schedulestr, " ");
		strcatsave(&schedulestr, num2psstr(minimagetime));
		strcatsave(&schedulestr, " ");
		strcatsave(&schedulestr, num2psstr(number));
		strcatsave(&schedulestr, " task");
		if (tmp->next) {
			strcatsave(&schedulestr, " ");
		}

		tmp = tmp->next;
	}

	/* output header */
	result = strsave("%%!PS-Adobe-2.0 EPSF-2.0\n");
	strcatsave(&result, "%%Creator: selview V" VERSION " (");
	strcatsave(&result, progname);
	strcatsave(&result, ")\n%%Title: ");
	strcatsave(&result, title);
	pe = getpwuid(getuid());
	if (pe) {
		strcatsave(&result, "\n%%For: ");
		strcatsave(&result, pe->pw_name);
	}
	strcatsave(&result, "\n%%Comment: All rights reserved.\n");
	strcatsave(&result, "%%Comment: Postscript code copyright (C) 1998-2005 by Ferdy Hanssen.\n");
	strcatsave(&result, "%%BoundingBox: ");
	bbl = -5 - 9 * namefieldlen;
	bbb = -48 - 32 * number;
	bbr = 3 + 2 * (maximagetime - minimagetime) * zoom;
	bbt = -46;
	strcatsave(&result, num2psstr(bbl));
	strcatsave(&result, " ");
	strcatsave(&result, num2psstr(bbb));
	strcatsave(&result, " ");
	strcatsave(&result, num2psstr(bbr));
	strcatsave(&result, " ");
	strcatsave(&result, num2psstr(bbt));
	strcatsave(&result, "\n%%EndComments\n%%EndProlog\n");

	strcatsave(&result, "/ZOOM ");
	strcatsave(&result, num2psstr(zoom));
	strcatsave(&result, " def\n/USELUCIDA false def\n");

	/* output functions */
	strcatsave(&result, epslibrary);
	strcatsave(&result, " ");

	/* output schedules */
	strcatsave(&result, schedulestr);
	memfree(schedulestr);
	strcatsave(&result, "\n");

	free_imagelist(schedules);

#ifdef DEBUG
	fprintf(stderr, "build_eps: end\n");
#endif

	return result;
}
