/*
 *  sel2*, selview: several programs to view schedule event listings
 *  Copyright (C) 1998-2005  Ferdy Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>

#include <gd.h>

/* The font to use for text in the picture */

#include <gdfonts.h>
#define CHAR_WIDTH 6
#define CHAR_HEIGHT 13
#define FONT gdFontSmall

/* length of string to use for creating numbers, hopefully long enough */
#define NUMSTRINGLEN 64

#include "actions.h"
#include "error.h"
#include "gdCrop.h"
#include "gdImageArrow.h"
#include "memalloc.h"
#include "output_graph.h"
#include "sel.h"
#include "tasks.h"

extern tasktree *tasks;

typedef struct vertArrowlist {
	int x;
	int y1;
	int y2;
	struct vertArrowlist *next;
} vertArrowlist;

typedef struct barlist {
	int x1;
	int x2;
	int y1;
	int y2;
	int id; /* only used to denote the resource */
	struct barlist *next;
} barlist;

typedef struct imagelist {
	gdImagePtr im;
	struct imagelist *next;
} imagelist;

/* This macro uses the LOCAL variables offset, minimagetime, and zoom */
#define	TIME2COOR(x)	(int) (offset + ((x) - minimagetime) * zoom)

typedef struct color {
	int r, g, b;
} color;

#define MAX_COLORS	12

color colorpool[MAX_COLORS] = {
/* This table may not contain black or any greys, those are used elsewhere */
/*	  R , G , B */
	{255,101,101}, /* pale red */
	{  0,255,  0}, /* bright green */
	{  0,111,255}, /* light blue */
	{255,255,  0}, /* yellow */
	{255,  0,255}, /* magenta */
	{  0,255,255}, /* cyan */
	{255,128,  0}, /* orange */
	{188, 92,179}, /* light purple */
	{ 24,228,108}, /* soft green */
	{181, 83, 26}, /* brown */
	{135,206,250}, /* light sky blue */
	{146,137,111}  /* army greenish */
};
color bg_color = {215,215,215}; /* light grey */
color fg_color = {  0,  0,  0}; /* black */
static int curr_color;

#define PATTERN_WIDTH	4
#define PATTERN_HEIGHT	4

typedef int pattern[PATTERN_HEIGHT][PATTERN_WIDTH];

#define MAX_PATTERNS	4

pattern patternpool[MAX_PATTERNS] = {
/* This table contains the patterns used to draw the different resources.
 * Because of the limited number of patterns it is possible that several
 * resources are drawn using the same pattern.
 * The values included are either 0 (for transparent) or 1 (for black).
 */
	{{1,1,0,0},  /* Pattern 1: diagonal stripes from NW to SE */
	 {0,1,1,0},
	 {0,0,1,1},
	 {1,0,0,1}},

	{{0,0,1,1},  /* Pattern 2: diagonal stripes from NE to SW */
	 {0,1,1,0},
	 {1,1,0,0},
	 {1,0,0,1}},

	{{1,1,0,0},  /* Pattern 3: checkers */
	 {1,1,0,0},
	 {0,0,1,1},
	 {0,0,1,1}},

	{{0,0,0,0},  /* Pattern 4: horizontal stripes */
	 {1,1,1,1},
	 {1,1,1,1},
	 {0,0,0,0}}
};

static int get_new_color(gdImagePtr im)
{
	int col;

	col = gdImageColorExact(im, colorpool[curr_color].r, colorpool[curr_color].g,
	                        colorpool[curr_color].b);
	if (col == -1) {
		col = gdImageColorAllocate(im, colorpool[curr_color].r, colorpool[curr_color].g,
		                           colorpool[curr_color].b);
		if (col == -1) {
			col = gdImageColorClosest(im, colorpool[curr_color].r,
			                          colorpool[curr_color].g,
			                          colorpool[curr_color].b);
		}
	}
	curr_color++;
	if (curr_color == MAX_COLORS) {
		curr_color = 0;
	}
	return col;
}

static void add_vertArrow(vertArrowlist *val, int x, int y1, int y2)
{
	vertArrowlist *tmp, *elem;

	tmp = val; /* first element of val is bogus */
	while (tmp->next) { /* search last element of list */
		tmp = tmp->next;
	}
	elem = memalloc(sizeof(vertArrowlist));
	elem->x = x;
	elem->y1 = y1;
	elem->y2 = y2;
	elem->next = NULL;
	tmp->next = elem;
}

static void free_vertArrowlist(vertArrowlist *val)
{
	if (val) {
		free_vertArrowlist(val->next);
		memfree(val);
	}
}

static void add_bar(barlist *bl, int x1, int y1, int x2, int y2, int id)
{
	barlist *tmp, *elem;

	tmp = bl; /* first element of bl is bogus */
	while (tmp->next) { /* search last element of list */
		tmp = tmp->next;
	}
	elem = memalloc(sizeof(barlist));
	elem->x1 = x1;
	elem->y1 = y1;
	elem->x2 = x2;
	elem->y2 = y2;
	elem->id = id;
	elem->next = NULL;
	tmp->next = elem;
}

static void free_barlist(barlist *bl)
{
	if (bl) {
		free_barlist(bl->next);
		memfree(bl);
	}
}

static void free_imagelist(imagelist *il)
{
	if (il) {
		free_imagelist(il->next);
		gdImageDestroy(il->im);
		memfree(il);
	}
}

static void add_action_to_image(
	gdImagePtr im,
	actiontree *at,
	int offset,
	unsigned long minimagetime,
	unsigned long maximagetime,
	vertArrowlist *val,
	barlist *execlist,
	barlist *resourcelist)
{
	unsigned long start, end, correction;

	correction = 1;
	if (at) {
		add_action_to_image(im, at->sooner, offset, minimagetime, maximagetime, val,
		                    execlist, resourcelist);
		start = at->start_time;
		end = at->end_time;
		switch (at->type) {
		case period:
			if (start >= minimagetime && start <= maximagetime) {
				add_vertArrow(val, TIME2COOR(start), 12, 2);
			}
			if (end >= minimagetime && end <= maximagetime) {
				add_vertArrow(val, TIME2COOR(end), 2, 12);
			}
			break;
		case exec:
			if (end < minimagetime || start > maximagetime) {
				break;
			}
			if (start < minimagetime) {
				start = minimagetime;
				correction = 0;
			}
			if (end > maximagetime) {
				end = maximagetime;
			}
			add_bar(execlist, TIME2COOR(start) + correction, 2, TIME2COOR(end), 12, 0);
			break;
		case resource_claim:
			if (end < minimagetime || start > maximagetime) {
				break;
			}
			if (start < minimagetime) {
				start = minimagetime;
				correction = 0;
			}
			if (end > maximagetime) {
				end = maximagetime;
			}
			add_bar(resourcelist, TIME2COOR(start) + correction, 2, TIME2COOR(end), 12,
			        at->rsrc_id);
			break;
		case deadline_redef:
			if (at->new_deadline >= minimagetime &&
			    at->new_deadline <= maximagetime) {
				add_vertArrow(val, TIME2COOR(at->new_deadline), 2, 9);
				add_vertArrow(val, TIME2COOR(at->new_deadline), 2, 12);
			}
			break;
		}
		add_action_to_image(im, at->later, offset, minimagetime, maximagetime, val,
		                    execlist, resourcelist);
	}
}

static gdImagePtr make_image_from_actions(
	actiontree *at,
	int imagewidth,
	int namefieldlen,
	char *name,
	int offset,
	unsigned long minimagetime,
	unsigned long maximagetime)
{
	gdImagePtr im;
	int fg, bg, bar;
	vertArrowlist *val;
	barlist *bars, *resources;
	gdImagePtr pattern[MAX_PATTERNS];
	int i, tile_darkgrey, tile_bar;
	char *namefield;

	im = gdImageCreate(imagewidth, 16); /* 2 empty lines above;
	                                     * 11 lines for the colored bar and
	                                     *              arrows;
	                                     * 1 line for the supporting line;
	                                     * 2 empty lines below
	                                     */
	bg = gdImageColorAllocate(im, bg_color.r, bg_color.g, bg_color.b);
	fg = gdImageColorAllocate(im, fg_color.r, fg_color.g, fg_color.b);
	bar = get_new_color(im);

	gdImageLine(im, offset, 13, TIME2COOR(maximagetime), 13, fg);

	val = memalloc(sizeof(vertArrowlist)); /* first element of list is bogus */
	val->x = val->y1 = val->y2 = 0;
	val->next = NULL;

	bars = memalloc(sizeof(barlist)); /* first element of list is bogus */
	bars->x1 = bars->x2 = bars->y1 = bars->y2 = bars->id = 0;
	bars->next = NULL;

	resources = memalloc(sizeof(barlist)); /* first element of list is bogus */
	resources->x1 = resources->x2 = resources->y1 = resources->y2 = resources->id = 0;
	resources->next = NULL;

	add_action_to_image(im, at, offset, minimagetime, maximagetime, val, bars, resources);

	if (bars->next) { /* draw bars */
		barlist *tmp;

		tmp = bars->next;
		while (tmp) {
			gdImageFilledRectangle(im, tmp->x1, tmp->y1, tmp->x2, tmp->y2, bar);
			tmp = tmp->next;
		}
	}

	for (i = 0; i < MAX_PATTERNS; i++) { /* create resource fill patterns */
		int x, y;

		pattern[i] = gdImageCreate(4, 4);
		/* get background color */
		tile_bar = gdImageColorAllocate(pattern[i], gdImageRed(im, bar),
		                                gdImageGreen(im, bar), gdImageBlue(im, bar));
		/* get dark grey */
		tile_darkgrey = gdImageColorAllocate(pattern[i], 80, 80, 80);
		for (y = 0; y < PATTERN_HEIGHT; y++) {
			for (x = 0; x < PATTERN_WIDTH; x++) {
				if (patternpool[i][y][x]) {
					gdImageSetPixel(pattern[i], x, y, tile_darkgrey);
				}
			}
		}
		gdImageColorTransparent(pattern[i], tile_bar);
	}

	if (resources->next) { /* draw resources */
		barlist *tmp;

		tmp = resources->next;
		while (tmp) {
			gdImageSetTile(im, pattern[(tmp->id) % MAX_PATTERNS]);
			gdImageFilledRectangle(im, tmp->x1, tmp->y1, tmp->x2, tmp->y2, gdTiled);
			tmp = tmp->next;
		}
	}

	for (i = 0; i < MAX_PATTERNS; i++) { /* clean up fill patterns */
		gdImageDestroy(pattern[i]);
	}

	if (val->next) { /* draw arrows */
		vertArrowlist *tmp;

		tmp = val->next;
		while (tmp) {
			gdImageVertArrow(im, tmp->x, tmp->y1, tmp->y2, fg, 3);
			tmp = tmp->next;
		}
	}

	/* clean up todo lists */
	free_barlist(resources);
	free_barlist(bars);
	free_vertArrowlist(val);

	/* draw task name */
	namefield = memalloc((size_t) (namefieldlen + 1));
	snprintf(namefield, (size_t) (namefieldlen + 1), "%*s", namefieldlen, name);
	gdImageString(im, FONT, 0, 2, namefield, fg);
	memfree(namefield);

	return im;
}

static imagelist *make_image_from_task(
	tasktree *task,
	int imagewidth,
	int namefieldlen,
	int offset,
	unsigned long minimagetime,
	unsigned long maximagetime)
{
	imagelist *result, *leftlast, *tmp;
	gdImagePtr im;

	result = NULL;
	if (task) {
		/* get left part */
		result = make_image_from_task(task->left, imagewidth, namefieldlen, offset,
		                              minimagetime, maximagetime);

		/* get my image */
		im = make_image_from_actions(task->actions, imagewidth, namefieldlen,
		                             task->name, offset, minimagetime, maximagetime);
		tmp = memalloc(sizeof(imagelist));
		tmp->im = im;
		tmp->next = NULL;

		if (result) {
			/* search for last element of left part */
			leftlast = result;
			while (leftlast->next) {
				leftlast = leftlast->next;
			}

			/* put my image at the back of the left part */
			leftlast->next = tmp;
			leftlast = leftlast->next;
		} else {
			result = leftlast = tmp;
		}

		/* get right part */
		tmp = make_image_from_task(task->right, imagewidth, namefieldlen, offset,
		                           minimagetime, maximagetime);

		/* concatenate left and right parts */
		leftlast->next = tmp;
	}
	return result;
}

gdImagePtr build_image(desttype destination)
{
	int namefieldlen, imagewidth, offset, bg, fg, height;
	unsigned long minimagetime, maximagetime;
	unsigned long mintimebartime, maxtimebartime;
	gdImagePtr timebar, picture, bordered;
	imagelist *schedules, *tmp;
	unsigned long i;
	int barx, length, step, numstep;
	char s[NUMSTRINGLEN], *namefield;

#ifdef DEBUG
	fprintf(stderr, "build_image: start\n");
#endif

	curr_color = 0;
	namefieldlen = get_longest_name_length_from_tasks();
	if (namefieldlen < 5) { /* 5 is the length of the word Tasks */
		namefieldlen = 5;
	}
	offset = (namefieldlen + 1) * CHAR_WIDTH + 1;
	minimagetime = (get_lowest_action_time_from_tasks() / 50) * 50;
	if (minimagetime < mintime) {
		minimagetime = mintime;
	} else if (minimagetime > maxtime) {
		minimagetime = maxtime;
	}
	maximagetime = ((get_highest_action_time_from_tasks() + 49) / 50) * 50;
	if (maximagetime > maxtime) {
		maximagetime = maxtime;
	} else if (maximagetime < mintime) {
		maximagetime = mintime;
	}
	if (maximagetime <= minimagetime) {
		maximagetime = minimagetime + 50;
	}
	imagewidth = (maximagetime - minimagetime + 1) * zoom + offset + CHAR_WIDTH;
	if (imagewidth > 32767) {
#ifdef OUTPUT_GUI
		selerror("picture will be wider than 32767 pixels");
		exit(1);
#else
		selwarning("picture produced will be wider than 32767 pixels");
#endif
	}
	{
		char s[NUMSTRINGLEN];

		if (snprintf(s, NUMSTRINGLEN, "%lu", maximagetime) == -1) {
			selerror("highest action time too high to produce output");
			return NULL;
		}
	}

	/*
	 * The final image will be composed of several images on top of
	 * each other, where the top image will contain the time bar, and
	 * the images underneath will each contain the schedule of one task.
	 */
	timebar = gdImageCreate(imagewidth, CHAR_HEIGHT + 13); /* time bar image */
	bg = gdImageColorAllocate(timebar, bg_color.r, bg_color.g, bg_color.b);
	fg = gdImageColorAllocate(timebar, fg_color.r, fg_color.g, fg_color.b);

	/* draw the time bar */
	mintimebartime = minimagetime;
	maxtimebartime = maximagetime;
	barx = offset;
	step = 5;
	if (zoom >= 3) {
		step = 1;
	} else {
		mintimebartime = ((minimagetime + 4) / 5) * 5;
		if (mintimebartime != minimagetime) {
			barx += (5 - (minimagetime % 5)) * zoom;
		}
	}
	numstep = 50;
	if (zoom >= 5) {
		numstep = 10;
	}
	if (zoom >= 10) {
		numstep = 5;
	}
	for (i = mintimebartime; i <= maxtimebartime; i += step, barx += step * zoom) {
		length = 1;
		if (i % 5 == 0) {
			length = 4;
			if (i % 10 == 0) {
				length = 8;
				if (i % 50 == 0) {
					length = 12;
				}
			}
		}
		gdImageLine(timebar, barx, CHAR_HEIGHT + 12 - length, barx,
			    CHAR_HEIGHT + 12, fg);
		if (i % numstep == 0) {
			snprintf(s, NUMSTRINGLEN, "%lu", i);
			gdImageString(timebar, FONT,
				      (int) (barx - strlen(s) * CHAR_WIDTH + CHAR_WIDTH / 2),
				      0, s, fg);
		}
	}
	namefield = memalloc((size_t) (namefieldlen + 1));
	snprintf(namefield, (size_t) (namefieldlen + 1), "%*s", namefieldlen, "Tasks");
	gdImageString(timebar, FONT, 0, 12, namefield, fg);
	gdImageLine(timebar, 0, CHAR_HEIGHT + 12, namefieldlen * CHAR_WIDTH - 1,
	            CHAR_HEIGHT + 12, fg);
	memfree(namefield);

	/* make schedule images */
	schedules = make_image_from_task(tasks, imagewidth, namefieldlen, offset, minimagetime,
	                                 maximagetime);

	/* paste all images together into one large image */
	height = gdImageSY(timebar);
	tmp = schedules;
	while (tmp) {
		height += gdImageSY(tmp->im);
		tmp = tmp->next;
	}
	picture = gdImageCreate(imagewidth, height);
	gdImageCopy(picture, timebar, 0, 0, 0, 0, gdImageSX(timebar), gdImageSY(timebar));
	height = gdImageSY(timebar);
	tmp = schedules;
	while (tmp) {
		gdImageCopy(picture, tmp->im, 0, height, 0, 0, gdImageSX(tmp->im),
		            gdImageSY(tmp->im));
		height += gdImageSY(tmp->im);
		tmp = tmp->next;
	}
	gdImageDestroy(timebar);

	free_imagelist(schedules);

	/* first crop it, then add a 3 pixel border */
	gdCrop(&picture, bg);
	bordered = gdImageCreate(gdImageSX(picture) + 6, gdImageSY(picture) + 6);
	gdImageCopy(bordered, picture, 3, 3, 0, 0, gdImageSX(picture), gdImageSY(picture));
	gdImageDestroy(picture);
	picture = bordered;

	if (transparent && destination == file) {
		gdImageColorTransparent(picture, bg);
	}

#ifdef DEBUG
	fprintf(stderr, "build_image: end\n");
#endif

	return picture;
}
