/*
 *  sel2*, selview: several programs to view schedule event listings
 *  Copyright (C) 1998-2005  Ferdy Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <limits.h>

#define OUTPUT_NUM

#include "actions.h"
#include "actions_num.h"
#include "error.h"
#include "memalloc.h"
#include "sel.h"
#include "tasks.h"

extern tasktree *tasks;
extern resourcetree *resources;

static int highest_rsrc_id_plus_one;

static unsigned long grand_ttl_exec_time;

static void show_task(
	tasktree *node,
	FILE *out,
	unsigned long **ttl_rsrc_usage_per_rsrc,
	int **rsrc_used_per_rsrc);
static void show_resource(
	resourcetree *node,
	FILE *out,
	unsigned long *ttl_rsrc_usage_per_rsrc,
	int *rsrc_used_per_rsrc);
static void show_action(
	actionlist *al,
	unsigned long *ttl_exec_time,
	unsigned long *num_of_active_periods,
	unsigned long *num_of_periods,
	unsigned long **rsrc_usage,
	unsigned long **num_of_rsrc_usages,
	int **rsrc_used,
	unsigned long *ttl_delay_time,
	FILE *out);
static void show_action_rec(
	actionlist *al,
	unsigned long *ttl_exec_time,
	unsigned long *num_of_active_periods,
	unsigned long *num_of_periods,
	int *in_period,
	int *in_exec,
	int *had_exec,
	unsigned long *start_exec_time,
	unsigned long **rsrc_usage,
	unsigned long **num_of_rsrc_usages,
	int **rsrc_used,
	int **in_rsrc,
	unsigned long **rsrc_usage_start_time,
	unsigned long *curr_release_time,
	unsigned long *first_exec_in_period,
	unsigned long *ttl_delay_time,
	unsigned long **rsrc_usage_overshoot,
	FILE *out);

void show(FILE *out)
{
	int i;
	unsigned long *ttl_rsrc_usage_per_rsrc;
	int *rsrc_used_per_rsrc;

	if (data) {
		fprintf(out, "%u\t# number of tasks\n", number_of_tasks);
	} else {
		fprintf(out, "Total number of tasks: %u\n", number_of_tasks);
		if (maxtime != ((ULONG_MAX - 50) / 1000) * 1000) {
			fprintf(out, "Discarding actions with a timestamp larger than %lu\n",
			             maxtime);
		}
	}
	highest_rsrc_id_plus_one = get_highest_rsrc_id() + 1;
	grand_ttl_exec_time = 0;
	ttl_rsrc_usage_per_rsrc = memalloc(sizeof(unsigned long) * highest_rsrc_id_plus_one);
	rsrc_used_per_rsrc = memalloc(sizeof(int) * highest_rsrc_id_plus_one);
	for (i = 0; i < highest_rsrc_id_plus_one; i++) {
		ttl_rsrc_usage_per_rsrc[i] = 0;
		rsrc_used_per_rsrc[i] = 0;
	}
	show_task(tasks, out, &ttl_rsrc_usage_per_rsrc,
	          &rsrc_used_per_rsrc);
	if (!data) {
		show_resource(resources, out, ttl_rsrc_usage_per_rsrc, rsrc_used_per_rsrc);
# ifdef DEBUG
		fprintf(out, "Longest length of any task's name: %lu\n",
		             get_longest_name_length_from_tasks());
# endif
		fprintf(out, "Highest action time of any task: %lu\n",
		             get_highest_action_time_from_tasks());
		fprintf(out, "Grand total execution time of all tasks: %lu\n",
		             grand_ttl_exec_time);
	}
	memfree(rsrc_used_per_rsrc);
	memfree(ttl_rsrc_usage_per_rsrc);
}

static void show_task(
	tasktree *node,
	FILE *out,
	unsigned long **ttl_rsrc_usage_per_rsrc,
	int **rsrc_used_per_rsrc)
{
	actionlist *al;
	unsigned long ttl_exec_time, num_of_active_periods, num_of_periods;
	unsigned long *rsrc_usage, *num_of_rsrc_usages;
	int *rsrc_used;
	unsigned long ttl_rsrc_usage, ttl_delay_time;
	unsigned long num_of_rsrcs_used;
	int i;

	ttl_exec_time = 0;
	num_of_active_periods = 0;
	num_of_periods = 0;
	rsrc_usage = memalloc(sizeof(unsigned long) * highest_rsrc_id_plus_one);
	num_of_rsrc_usages = memalloc(sizeof(unsigned long) * highest_rsrc_id_plus_one);
	rsrc_used = memalloc(sizeof(int) * highest_rsrc_id_plus_one);
	for (i = 0; i < highest_rsrc_id_plus_one; i++) {
		rsrc_usage[i] = 0;
		num_of_rsrc_usages[i] = 0;
		rsrc_used[i] = 0;
	}
	ttl_delay_time = 0;
	if (node) {
		show_task(node->left, out, ttl_rsrc_usage_per_rsrc, rsrc_used_per_rsrc);
		if (data) {
			fprintf(out, "%d\t# task id\n", node->id);
		} else {
			fprintf(out, "  Task %2d \"%s\"\n", node->id, node->name);
		}
		al = convert_tree_to_list(node->actions);
		show_action(al, &ttl_exec_time, &num_of_active_periods, &num_of_periods,
		            &rsrc_usage, &num_of_rsrc_usages, &rsrc_used, &ttl_delay_time, out);
		free_actionlist(al);
		num_of_rsrcs_used = 0;
		for (i = 0; i < highest_rsrc_id_plus_one; i++) {
			if (rsrc_used[i]) {
				num_of_rsrcs_used++;
			}
		}
		if (data) {
			fprintf(out, "%lu\t# total execution time\n", ttl_exec_time);
			fprintf(out, "%lu\t# total time between release and exec\n",
			             ttl_delay_time);
			fprintf(out, "%lu\t# number of active periods\n", num_of_active_periods);
			fprintf(out, "%lu\t# number of periods\n", num_of_periods);
			fprintf(out, "%lu\t# number of resources used by this task\n",
			             num_of_rsrcs_used);
		} else {
			fprintf(out, "    Highest action time: %lu\n",
			             get_highest_time_from_actions(node->actions));
			fprintf(out, "    Total execution time: %lu\n", ttl_exec_time);
			fprintf(out, "    Total time between start of period and execution: "
			             "%lu\n", ttl_delay_time);
			fprintf(out, "    Total number of active periods: %lu\n",
			             num_of_active_periods);
			fprintf(out, "    Total number of periods: %lu\n", num_of_periods);
		}
		if (num_of_rsrcs_used) {
			if (!data) {
				fprintf(out, "    Resource usage per resource:\n");
			}
			ttl_rsrc_usage = 0;
			for (i = 0; i < highest_rsrc_id_plus_one; i++) {
				if (rsrc_used[i]) {
					if (data) {
						fprintf(out, "%d\t# resource id\n"
						             "%lu\t# total resource usage time\n"
						             "%lu\t# times resource used\n", i,
						             rsrc_usage[i],
						             num_of_rsrc_usages[i]);
					} else {
						fprintf(out, "      Resource %2d: %lu total "
						             "time, %lu times used, %.2f "
						             "average per usage\n", i,
						             rsrc_usage[i],
						             num_of_rsrc_usages[i],
						             (double) rsrc_usage[i] / (double) num_of_rsrc_usages[i]);
						ttl_rsrc_usage += rsrc_usage[i];
					}
					(*ttl_rsrc_usage_per_rsrc)[i] += rsrc_usage[i];
					(*rsrc_used_per_rsrc)[i] += num_of_rsrc_usages[i];
				}
			}
			if (!data) {
				fprintf(out, "    Total resource usage: %lu\n", ttl_rsrc_usage);
			}
		}
		if (!data) {
			fprintf(out, "    Average time between start of period and execution: "
			             "%.2f\n",
			             (double) ttl_delay_time / (double) num_of_active_periods);
			fprintf(out, "    Average execution time per period: %.2f\n",
			             (double) ttl_exec_time / (double) num_of_active_periods);
			grand_ttl_exec_time += ttl_exec_time;
		}
		show_task(node->right, out, ttl_rsrc_usage_per_rsrc, rsrc_used_per_rsrc);
	}
	memfree(rsrc_used);
	memfree(num_of_rsrc_usages);
	memfree(rsrc_usage);
}

static void show_resource(
	resourcetree *node,
	FILE *out,
	unsigned long *ttl_rsrc_usage_per_rsrc,
	int *rsrc_used_per_rsrc)
{
	if (node) {
		show_resource(node->left, out, ttl_rsrc_usage_per_rsrc, rsrc_used_per_rsrc);
		fprintf(out, "  Resource %2d \"%s\"\n", node->id, node->name);
		fprintf(out, "    Total time used: %lu\n", ttl_rsrc_usage_per_rsrc[node->id]);
		fprintf(out, "    Number of times used: %d\n", rsrc_used_per_rsrc[node->id]);
		fprintf(out, "    Average usage time per claim: %.2f\n",
		             (double) ttl_rsrc_usage_per_rsrc[node->id] / (double) rsrc_used_per_rsrc[node->id]);
		show_resource(node->right, out, ttl_rsrc_usage_per_rsrc, rsrc_used_per_rsrc);
	}
}

static void show_action(
	actionlist *al,
	unsigned long *ttl_exec_time,
	unsigned long *num_of_active_periods,
	unsigned long *num_of_periods,
	unsigned long **rsrc_usage,
	unsigned long **num_of_rsrc_usages,
	int **rsrc_used,
	unsigned long *ttl_delay_time,
	FILE *out)
{
	int in_period, in_exec, had_exec;
	unsigned long start_exec_time;
	int *in_rsrc;
	unsigned long *rsrc_usage_start_time;
	unsigned long curr_release_time, first_exec_in_period, *rsrc_usage_overshoot;
	int i;

	in_period = 0;
	in_exec = 0;
	had_exec = 0;
	start_exec_time = 0;
	in_rsrc = memalloc(sizeof(int) * highest_rsrc_id_plus_one);
	rsrc_usage_start_time = memalloc(sizeof(unsigned long) * highest_rsrc_id_plus_one);
	rsrc_usage_overshoot = memalloc(sizeof(unsigned long) * highest_rsrc_id_plus_one);
	for (i = 0; i < highest_rsrc_id_plus_one; i++) {
		in_rsrc[i] = 0;
		rsrc_usage_start_time[i] = 0;
		rsrc_usage_overshoot[i] = 0;
	}
	curr_release_time = 0;
	first_exec_in_period = 1;
	show_action_rec(al, ttl_exec_time, num_of_active_periods, num_of_periods, &in_period,
	                &in_exec, &had_exec, &start_exec_time, rsrc_usage,
	                num_of_rsrc_usages, rsrc_used, &in_rsrc,
	                &rsrc_usage_start_time, &curr_release_time, &first_exec_in_period,
	                ttl_delay_time, &rsrc_usage_overshoot, out);
	memfree(rsrc_usage_overshoot);
	memfree(rsrc_usage_start_time);
	memfree(in_rsrc);
}

static void action_tree_corrupt(unsigned long time, const char *message)
{
	selerror("action tree corrupt (time = %lu): %s", time, message);
}

static void show_action_rec(
	actionlist *al,
	unsigned long *ttl_exec_time,
	unsigned long *num_of_active_periods,
	unsigned long *num_of_periods,
	int *in_period,
	int *in_exec,
	int *had_exec,
	unsigned long *start_exec_time,
	unsigned long **rsrc_usage,
	unsigned long **num_of_rsrc_usages,
	int **rsrc_used,
	int **in_rsrc,
	unsigned long **rsrc_usage_start_time,
	unsigned long *curr_release_time,
	unsigned long *first_exec_in_period,
	unsigned long *ttl_delay_time,
	unsigned long **rsrc_usage_overshoot,
	FILE *out)
{
	int i;

	if (al) {
		switch (al->type) {
		case release:
			if (*in_period) {
				action_tree_corrupt(al->time, "2 consecutive releases without "
				                              "intervening deadline");
			}
			*in_period = 1;
			*curr_release_time = al->time;
			*first_exec_in_period = 1;
			if (verbosity) {
				fprintf(out, "      Action: %s at %lu\n",
				             actionlisttype2ascii(al->type), al->time);
			}
			break;
		case deadline:
			if (!*in_period) {
				action_tree_corrupt(al->time, "deadline found without preceding "
				                              "release");
			} else {
				if (*had_exec) {
					(*num_of_active_periods)++;
					*had_exec = 0;
				}
				(*num_of_periods)++;
			}
			*in_period = 0;
			if (verbosity) {
				fprintf(out, "      Action: %s at %lu\n",
				             actionlisttype2ascii(al->type), al->time);
			}
			break;
		case exec_start:
			if (*in_exec) {
				action_tree_corrupt(al->time, "2 consecutive execution starts "
				                              "without intervening end");
			}
			*in_exec = 1;
			*start_exec_time = al->time;
			if (*first_exec_in_period) {
				if (verbosity) {
					fprintf(out, "      Time between release and exec: "
					             "%lu\n",
					             *start_exec_time - *curr_release_time);
				}
				*ttl_delay_time += *start_exec_time - *curr_release_time;
				*first_exec_in_period = 0;
			}
			for (i = 0; i < highest_rsrc_id_plus_one; i++) {
				if ((*in_rsrc)[i]) {
					(*rsrc_usage_overshoot)[i] = al->time - (*rsrc_usage_overshoot)[i];
					(*rsrc_usage)[i] -= (*rsrc_usage_overshoot)[i];
					if (verbosity) {
						fprintf(out, "      Subtracted %lu from "
						             "resource usage of resource %d\n",
						             (*rsrc_usage_overshoot)[i], i);
					}
					(*rsrc_usage_overshoot)[i] = 0;
				}
			}
			if (verbosity) {
				fprintf(out, "      Action: %s at %lu\n",
				             actionlisttype2ascii(al->type), al->time);
			}
			break;
		case exec_end:
			if (!*in_exec) {
				action_tree_corrupt(al->time, "execution end found without "
				                              "preceding start");
			} else {
				(*ttl_exec_time) += al->time - *start_exec_time;
				*had_exec = 1;
			}
			*in_exec = 0;
			*start_exec_time = 0;
			for (i = 0; i < highest_rsrc_id_plus_one; i++) {
				if ((*in_rsrc)[i]) {
					(*rsrc_usage_overshoot)[i] = al->time;
				}
			}
			if (verbosity) {
				fprintf(out, "      Action: %s at %lu\n",
				             actionlisttype2ascii(al->type), al->time);
			}
			break;
		case claim_res:
			if ((*in_rsrc)[al->rsrc_id]) {
				action_tree_corrupt(al->time, "2 consecutive resource claims "
				                              "without intervening free");
			}
			(*in_rsrc)[al->rsrc_id] = 1;
			(*rsrc_usage_start_time)[al->rsrc_id] = al->time;
			(*rsrc_used)[al->rsrc_id] = 1;
			if (verbosity) {
				fprintf(out, "      Action: %s %d at %lu\n",
				             actionlisttype2ascii(al->type), al->rsrc_id,
				             al->time);
			}
			break;
		case free_res:
			if (!(*in_rsrc)[al->rsrc_id]) {
				action_tree_corrupt(al->time, "resource free found without "
				                              "preceding claim");
			} else {
				(*rsrc_usage)[al->rsrc_id] += al->time - (*rsrc_usage_start_time)[al->rsrc_id];
				((*num_of_rsrc_usages)[al->rsrc_id])++;
			}
			(*in_rsrc)[al->rsrc_id] = 0;
			(*rsrc_usage_start_time)[al->rsrc_id] = 0;
			if (verbosity) {
				fprintf(out, "      Action: %s %d at %lu\n",
				             actionlisttype2ascii(al->type), al->rsrc_id,
				             al->time);
			}
			break;
		case redef_deadline:
			if (verbosity) {
				fprintf(out, "      Action: %s to %lu at %lu\n",
				             actionlisttype2ascii(al->type), al->new_deadline,
				             al->time);
			}
			break;
		default:
			action_tree_corrupt(al->time, "unknown action");
		}
		show_action_rec(al->next, ttl_exec_time, num_of_active_periods, num_of_periods,
		                in_period, in_exec, had_exec, start_exec_time, rsrc_usage,
		                num_of_rsrc_usages, rsrc_used, in_rsrc,
		                rsrc_usage_start_time, curr_release_time,
		                first_exec_in_period, ttl_delay_time, rsrc_usage_overshoot,
		                out);
	}
}
