/*
 *  sel2*, selview: several programs to view schedule event listings
 *  Copyright (C) 1998-2005  Ferdy Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* This file contains a replacement for the standard yyerror() function */

#include <stdio.h>
#include <stdarg.h>

#ifdef OUTPUT_GUI
# include <gtk/gtk.h>
#endif

#include "error.h"
#ifdef OUTPUT_GUI
# include "gui.h"
#endif
#include "sel.h"

#ifdef OUTPUT_GUI
#define MAX_ERROR_LENGTH	256

static int gui_error_init_done = 0;
static int error_window_shown;
static GtkWidget *error_window, *error_text;
#endif

void selerror(const char *s, ...) /* yyerror */
/*
 * The string s is a format string, similar to printf.
 * The output will be: <progname>: error in line <sellineno>: s ...
 * Please note that it will ALWAYS print an extra newline.
 */
{
	va_list ap;
#ifdef OUTPUT_GUI
	char err[MAX_ERROR_LENGTH];
#endif

	va_start(ap, s);
#ifdef OUTPUT_GUI
	if (gui_error_init_done) {
		gdk_window_set_icon(error_window->window, NULL, icon, NULL);
		gtk_widget_show(error_window);
		error_window_shown = 1;
		if (sellineno) {
			snprintf(err, MAX_ERROR_LENGTH, "error in line %d: ", sellineno);
		} else {
			snprintf(err, MAX_ERROR_LENGTH, "error: ");
		}
		gtk_text_insert(GTK_TEXT(error_text), NULL, NULL, NULL, err, -1);
		vsnprintf(err, MAX_ERROR_LENGTH, s, ap);
		gtk_text_insert(GTK_TEXT(error_text), NULL, NULL, NULL, err, -1);
		gtk_text_insert(GTK_TEXT(error_text), NULL, NULL, NULL, "\n", -1);
	} else {
#endif
		if (sellineno) {
			fprintf(stderr, "%s: error in line %d: ", progname, sellineno);
		} else {
			fprintf(stderr, "%s: error: ", progname);
		}
		vfprintf(stderr, s, ap);
		fprintf(stderr, "\n");
#ifdef OUTPUT_GUI
	}
#endif
	va_end(ap);
}

void selwarning(const char *s, ...) /* yywarning */
/*
 * The string s is a format string, similar to printf.
 * The output will be: <progname>: warning in line <sellineno>: s ...
 * Please note that it will ALWAYS print an extra newline.
 */
{
	va_list ap;
#ifdef OUTPUT_GUI
	char warn[MAX_ERROR_LENGTH + 1];
#endif

	va_start(ap, s);
#ifdef OUTPUT_GUI
	if (gui_error_init_done) {
		gdk_window_set_icon(error_window->window, NULL, icon, NULL);
		gtk_widget_show(error_window);
		error_window_shown = 1;
		if (sellineno) {
			snprintf(warn, MAX_ERROR_LENGTH, "warning in line %d: ", sellineno);
		} else {
			snprintf(warn, MAX_ERROR_LENGTH, "warning: ");
		}
		gtk_text_insert(GTK_TEXT(error_text), NULL, NULL, NULL, warn, -1);
		vsnprintf(warn, MAX_ERROR_LENGTH, s, ap);
		gtk_text_insert(GTK_TEXT(error_text), NULL, NULL, NULL, warn, -1);
		gtk_text_insert(GTK_TEXT(error_text), NULL, NULL, NULL, "\n", -1);
	} else {
#endif
		if (sellineno) {
			fprintf(stderr, "%s: warning in line %d: ", progname, sellineno);
		} else {
			fprintf(stderr, "%s: warning: ", progname);
		}
		vfprintf(stderr, s, ap);
		fprintf(stderr, "\n");
#ifdef OUTPUT_GUI
	}
#endif
	va_end(ap);
}

#ifdef OUTPUT_GUI
void selinfo(const char *s, ...) /* yyinfo */
/*
 * The string s is a format string, similar to printf.
 * The output will be: s ...
 * Please note that it will ALWAYS print an extra newline.
 */
{
	va_list ap;
	char info[MAX_ERROR_LENGTH + 1];

	va_start(ap, s);
	vsnprintf(info, MAX_ERROR_LENGTH, s, ap);
	gtk_text_insert(GTK_TEXT(error_text), NULL, NULL, NULL, info, -1);
	gtk_text_insert(GTK_TEXT(error_text), NULL, NULL, NULL, "\n", -1);
	va_end(ap);
}

static void hide_error_window(GtkWidget *w, GtkWidget *error_window)
{
	unused(w);

	if (error_window_shown) {
		gtk_widget_hide(error_window);
		error_window_shown = 0;
	}
}

void gui_error_init(void)
{
	GtkWidget *vbox, *bbox, *textbox, *textscroll, *ok_button;

	error_window = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_window_set_wmclass(GTK_WINDOW(error_window), "messages", "Selview");
	gtk_window_set_title(GTK_WINDOW(error_window), "selview: messages");
	gtk_widget_realize(error_window);
	gtk_window_set_position(GTK_WINDOW(error_window), GTK_WIN_POS_CENTER);
	gtk_signal_connect(GTK_OBJECT(error_window), "delete_event",
	                   GTK_SIGNAL_FUNC(hide_error_window), error_window);
	gtk_quit_add_destroy(1, GTK_OBJECT(error_window));
	gtk_container_set_border_width(GTK_CONTAINER(error_window), 10);

	vbox = gtk_vbox_new(FALSE, 10);
	gtk_container_add(GTK_CONTAINER(error_window), vbox);
	gtk_widget_show(vbox);

	textbox = gtk_hbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(vbox), textbox);
	gtk_widget_show(textbox);

	error_text = gtk_text_new(NULL, NULL);
	gtk_text_set_editable(GTK_TEXT(error_text), FALSE);
	gtk_widget_set_usize(error_text, 450, 180);
	gtk_container_add(GTK_CONTAINER(textbox), error_text);
	gtk_widget_realize(error_text);
	gtk_widget_show(error_text);

	textscroll = gtk_vscrollbar_new(GTK_TEXT(error_text)->vadj);
	gtk_box_pack_start(GTK_BOX(textbox), textscroll, FALSE, FALSE, 0);
	gtk_widget_show(textscroll);

	bbox = gtk_hbutton_box_new();
	gtk_button_box_set_layout(GTK_BUTTON_BOX(bbox), GTK_BUTTONBOX_SPREAD);
	gtk_button_box_set_spacing(GTK_BUTTON_BOX(bbox), 5);
	gtk_box_pack_end(GTK_BOX(vbox), bbox, FALSE, FALSE, 0);
	gtk_widget_show(bbox);

	ok_button = gtk_button_new_with_label("OK");
	GTK_WIDGET_SET_FLAGS(ok_button, GTK_CAN_DEFAULT);
	gtk_widget_grab_default(ok_button);
	gtk_signal_connect(GTK_OBJECT(ok_button), "clicked",
	                   GTK_SIGNAL_FUNC(hide_error_window), error_window);
	gtk_container_add(GTK_CONTAINER(bbox), ok_button);
	gtk_widget_show(ok_button);

	error_window_shown = 0;
	gui_error_init_done = 1;
}
#endif
