/*
 *  sel2*, selview: several programs to view schedule event listings
 *  Copyright (C) 1998-2005  Ferdy Hanssen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <limits.h>
#include <string.h>

#include "actions.h"
#include "error.h"
#include "memalloc.h"
#include "sel.h"
#include "strsave.h"
#include "tasks.h"

unsigned int number_of_tasks;
unsigned int number_of_resources;

tasktree *tasks = NULL;
resourcetree *resources = NULL;

static tasktree *find_next_free_task_node(tasktree *node, int id)
{
	if (node == NULL) {
		return NULL;
	}
	if (id == node->id) {
		selerror("task %d declared twice", id);
		return NULL;
	} else if (id < node->id) {
		if (node->left == NULL) {
			return node;
		}
		return find_next_free_task_node(node->left, id);
	} else {
		if (node->right == NULL) {
			return node;
		}
		return find_next_free_task_node(node->right, id);
	}
}

void new_task(int tasknumber, char *name)
{
	tasktree *tmp;

	tmp = memalloc(sizeof(tasktree));
	tmp->id = tasknumber;
	tmp->name = strsave(name);
	tmp->actions = NULL;
	tmp->left = NULL;
	tmp->right = NULL;
	if (tasks == NULL) {
		tasks = tmp;
		number_of_tasks = 1;
	} else {
		tasktree *node;

		node = find_next_free_task_node(tasks, tasknumber);
		if (node) {
			if (tasknumber < node->id) {
				node->left = tmp;
			} else {
				node->right = tmp;
			}
			number_of_tasks++;
		}
	}
}

static void free_task(tasktree *node)
{
	if (node) {
		free_task(node->left);
		free_task(node->right);
		memfree(node->name);
		free_actions(node->actions);
		memfree(node);
		number_of_tasks--;
	}
}

void free_tasks(void)
{
	free_task(tasks);
	tasks = NULL;
}

static tasktree *get_task(tasktree *node, int id)
{
	if (node == NULL) {
		selerror("task %d not defined", id);
		return NULL;
	}
	if (id == node->id) {
		return node;
	} else if (id < node->id) {
		return get_task(node->left, id);
	} else {
		return get_task(node->right, id);
	}
}

void add_action_to_task(action_type type, int tasknumber, ...)
{
	tasktree *task;
	va_list ap;
	unsigned long s, t, n;
	int r;

	task = get_task(tasks, tasknumber);
	if (task) {
		va_start(ap, tasknumber);
		switch (type) {
		case period:
		case exec:
			s = va_arg(ap, unsigned long);	/* start time */
			t = va_arg(ap, unsigned long);	/* end time */
			task->actions = add_action(task->actions, type, s, t);
			break;
		case resource_claim:
			r = va_arg(ap, int);		/* rsrc_id */
			s = va_arg(ap, unsigned long);	/* start time */
			t = va_arg(ap, unsigned long);	/* end time */
			task->actions = add_action(task->actions, type, r, s, t);
			break;
		case deadline_redef:
			t = va_arg(ap, unsigned long);	/* time */
			n = va_arg(ap, unsigned long);	/* new deadline */
			task->actions = add_action(task->actions, deadline_redef, t, n);
			break;
		}
		va_end(ap);
	}
}

static unsigned long get_highest_action_time_from_task(tasktree *node,
                                                       unsigned long high_until_now)
{
	unsigned long high_from_action, tmp;

	if (node) {
		high_from_action = get_highest_time_from_actions(node->actions);
		if (high_from_action > high_until_now) {
			high_until_now = high_from_action;
		}
		tmp = get_highest_action_time_from_task(node->left, high_until_now);
		if (tmp > high_until_now) {
			high_until_now = tmp;
		}
		tmp = get_highest_action_time_from_task(node->right, high_until_now);
		if (tmp > high_until_now) {
			high_until_now = tmp;
		}
	}
	return high_until_now;
}

unsigned long get_highest_action_time_from_tasks(void)
{
	return get_highest_action_time_from_task(tasks, 0);
}

static unsigned long get_lowest_action_time_from_task(tasktree *node,
                                                      unsigned long low_until_now)
{
	unsigned long low_from_action, tmp;

	if (node) {
		low_from_action = get_lowest_time_from_actions(node->actions);
		if (low_from_action < low_until_now) {
			low_until_now = low_from_action;
		}
		tmp = get_lowest_action_time_from_task(node->left, low_until_now);
		if (tmp < low_until_now) {
			low_until_now = tmp;
		}
		tmp = get_lowest_action_time_from_task(node->right, low_until_now);
		if (tmp < low_until_now) {
			low_until_now = tmp;
		}
	}
	return low_until_now;
}

unsigned long get_lowest_action_time_from_tasks(void)
{
	return get_lowest_action_time_from_task(tasks, ULONG_MAX);
}

static unsigned long get_longest_name_length_from_task(tasktree *node,
                                                       unsigned long high_until_now)
{
	unsigned long tmp;

	if (node) {
		tmp = strlen(node->name);
		if (tmp > high_until_now) {
			high_until_now = tmp;
		}
		tmp = get_longest_name_length_from_task(node->left, high_until_now);
		if (tmp > high_until_now) {
			high_until_now = tmp;
		}
		tmp = get_longest_name_length_from_task(node->right, high_until_now);
		if (tmp > high_until_now) {
			high_until_now = tmp;
		}
	}
	return high_until_now;
}

unsigned long get_longest_name_length_from_tasks(void)
{
	return get_longest_name_length_from_task(tasks, 0);
}

static resourcetree *find_next_free_resource_node(resourcetree *node, int id)
{
	if (node == NULL) {
		return NULL;
	}
	if (id == node->id) {
		selerror("resource %d declared twice", id);
		return NULL;
	} else if (id < node->id) {
		if (node->left == NULL) {
			return node;
		}
		return find_next_free_resource_node(node->left, id);
	} else {
		if (node->right == NULL) {
			return node;
		}
		return find_next_free_resource_node(node->right, id);
	}
}

void new_resource(int resourcenumber, char *name)
{
	resourcetree *tmp;

	tmp = memalloc(sizeof(resourcetree));
	tmp->id = resourcenumber;
	tmp->name = strsave(name);
	tmp->left = NULL;
	tmp->right = NULL;
	if (resources == NULL) {
		resources = tmp;
		number_of_resources = 1;
	} else {
		resourcetree *node;

		node = find_next_free_resource_node(resources, resourcenumber);
		if (node) {
			if (resourcenumber < node->id) {
				node->left = tmp;
			} else {
				node->right = tmp;
			}
			number_of_resources++;
		}
	}
}

static void free_resource_(resourcetree *node)
{
	if (node) {
		free_resource_(node->left);
		free_resource_(node->right);
		memfree(node->name);
		memfree(node);
		number_of_resources--;
	}
}

void free_resources(void)
{
	free_resource_(resources);
	resources = NULL;
}

int get_highest_rsrc_id_(resourcetree *node)
{
	int highest_id, tmp;

	if (node) {
		highest_id = node->id;
		tmp = get_highest_rsrc_id_(node->left);
		if (tmp > highest_id) {
			highest_id = tmp;
		}
		tmp = get_highest_rsrc_id_(node->right);
		if (tmp > highest_id) {
			highest_id = tmp;
		}
		return highest_id;
	} else {
		return 0;
	}
}

int get_highest_rsrc_id(void)
{
	return get_highest_rsrc_id_(resources);
}
